from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt


class AbstractPulse(ABC):
    @abstractmethod
    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.float64]:
        pass

    @abstractmethod
    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.float64]:
        pass

    @property
    @abstractmethod
    def support(self) -> tuple[float, float]:
        pass



from typing import Optional

import numpy as np
import numpy.typing as npt

from .._types import ArrayIntLike

# TODO: Rename binlist2int and int2binlist to something better.
# TODO: Vectorize those functions (e.g., axis=1).


def binlist2int(binlist: ArrayIntLike) -> int:
    r"""
    Converts a bit array to its integer representation (LSB first).

    Parameters:
        binlist: A list or array of $0$'s and $1$'s whose $i$-th element stands for the coefficient of $2^i$ in the binary representation of the output integer.

    Returns:
        integer: The integer representation of the input bit array.

    Examples:
        >>> komm.binlist2int([0, 0, 0, 0, 1])
        16

        >>> komm.binlist2int([0, 1, 0, 1, 1])
        26

        >>> komm.binlist2int([0, 1, 0, 1, 1, 0, 0, 0])
        26
    """
    return sum(1 << i for (i, b) in enumerate(binlist) if b != 0)


def int2binlist(integer: int, width: Optional[int] = None) -> npt.NDArray[np.int_]:
    r"""
    Converts an integer to its bit array representation (LSB first).

    Parameters:
        integer: The input integer. May be any nonnegative integer.
        width: If this parameter is specified, the output will be filled with zeros on the right so that its length will be the specified value.

    Returns:
        binlist (Array1D[int]): An array of $0$'s and $1$'s whose $i$-th element stands for the coefficient of $2^i$ in the binary representation of the input integer.

    Examples:
        >>> komm.int2binlist(16)
        array([0, 0, 0, 0, 1])

        >>> komm.int2binlist(26)
        array([0, 1, 0, 1, 1])

        >>> komm.int2binlist(26, width=8)
        array([0, 1, 0, 1, 1, 0, 0, 0])
    """
    if width is None:
        width = max(integer.bit_length(), 1)
    return np.array([(integer >> i) & 1 for i in range(width)], dtype=int)


def pack(bit_array: ArrayIntLike, width: int) -> npt.NDArray[np.int_]:
    r"""
    Packs a given bit array. Splits the bit array into groups of `width` bits and converts each group to its integer value.

    Parameters:
        bit_array: The input bit array.
        width: The width of each group.

    Returns:
        packed (Array1D[int]): The packed integer array.

    Examples:
        >>> komm.pack([0, 0, 0, 0, 1, 0, 1, 0, 1, 1], width=5)
        array([16, 26])

        >>> komm.pack([0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0], width=8)
        array([16, 26])
    """
    return np.apply_along_axis(binlist2int, 1, np.reshape(bit_array, shape=(-1, width)))


def unpack(bit_array: ArrayIntLike, width: int) -> npt.NDArray[np.int_]:
    r"""
    Unpacks a given integer array. Unpacks a given integer array by converting each integer to its bit array representation, using the specified `width` for each group, and concatenating the results.

    Parameters:
        bit_array: The input integer array.
        width: The width of each group.

    Returns:
        unpacked (Array1D[int]): The unpacked bit array.

    Examples:
        >>> komm.unpack([16, 26], width=5)
        array([0, 0, 0, 0, 1, 0, 1, 0, 1, 1])

        >>> komm.unpack([16, 26], width=8)
        array([0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0])
    """
    return np.ravel([int2binlist(i, width=width) for i in bit_array])



from attrs import field, frozen
from typing_extensions import Self


@frozen
class Integer:
    value: int = field(converter=int)

    @property
    def ambient(self) -> "Integers":
        return Integers()

    def __add__(self, other: Self) -> Self:
        return self.__class__(self.value + other.value)

    def __sub__(self, other: Self) -> Self:
        return self.__class__(self.value - other.value)

    def __neg__(self) -> Self:
        return self.__class__(-self.value)

    def __mul__(self, other: Self) -> Self:
        return self.__class__(self.value * other.value)

    def __rmul__(self, other: int) -> Self:
        return self.__class__(self.value * other)

    def __divmod__(self, other: Self) -> tuple[Self, Self]:
        quotient, remainder = divmod(self.value, other.value)
        return self.__class__(quotient), self.__class__(remainder)

    def __floordiv__(self, other: Self) -> Self:
        return self.__class__(self.value // other.value)

    def __mod__(self, other: Self) -> Self:
        return self.__class__(self.value % other.value)


class Integers:
    def __call__(self, value: int) -> Integer:
        return Integer(value)

    @property
    def zero(self) -> Integer:
        return Integer(0)

    @property
    def one(self) -> Integer:
        return Integer(1)



from typing import Callable, Literal, TypedDict

import numpy as np
import numpy.typing as npt

Decoder = Callable[..., npt.NDArray[np.int_ | np.float64]]


class BlockDecoderData(TypedDict):
    description: str
    decoder: Decoder
    type_in: Literal["soft", "hard"]
    type_out: Literal["soft", "hard"]
    target: Literal["codeword", "message"]


class RegistryBlockDecoder:
    _registry: dict[str, BlockDecoderData] = {}

    @classmethod
    def register(cls, method: str, data: BlockDecoderData) -> None:
        cls._registry[method] = data

    @classmethod
    def is_registered(cls, method: str) -> bool:
        return method in cls._registry

    @classmethod
    def get(cls, method: str) -> BlockDecoderData:
        return cls._registry[method]

    @classmethod
    def list(cls) -> list[str]:
        return list(cls._registry.keys())



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt

from .._types import ArrayIntLike


class AbstractBlockCode(ABC):
    @property
    @abstractmethod
    def length(self) -> int:
        pass

    @property
    @abstractmethod
    def dimension(self) -> int:
        pass

    @property
    @abstractmethod
    def redundancy(self) -> int:
        pass

    @abstractmethod
    def enc_mapping(self, u: ArrayIntLike) -> npt.NDArray[np.int_]:
        pass

    @abstractmethod
    def inv_enc_mapping(self, v: ArrayIntLike) -> npt.NDArray[np.int_]:
        pass

    @abstractmethod
    def chk_mapping(self, r: ArrayIntLike) -> npt.NDArray[np.int_]:
        pass



from functools import cached_property

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._types import ArrayIntLike
from .BlockCode import BlockCode


@frozen(kw_only=True, eq=False)
class SystematicBlockCode(BlockCode):
    r"""
    Systematic linear block code. A *systematic linear block code* is a [linear block code](/ref/BlockCode) in which the information bits can be found in predefined positions in the codeword, called the *information set* $\mathcal{K}$, which is a $k$-sublist of $[0 : n)$; the remaining positions are called the *parity set* $\mathcal{M}$, which is a $m$-sublist of $[0 : n)$. In this case, the generator matrix then has the property that the columns indexed by $\mathcal{K}$ are equal to $I_k$, and the columns indexed by $\mathcal{M}$ are equal to $P$. The check matrix has the property that the columns indexed by $\mathcal{M}$ are equal to $I_m$, and the columns indexed by $\mathcal{K}$ are equal to $P^\transpose$. The matrix $P \in \mathbb{B}^{k \times m}$ is called the *parity submatrix* of the code.

    The constructor expects the parity submatrix and the information set.

    Parameters:
        parity_submatrix (Array2D[int]): The parity submatrix $P$ the code, which is a $k \times m$ binary matrix.

        information_set (Optional[Array1D[int] | str]): Either an array containing the indices of the information positions, which must be a $k$-sublist of $[0 : n)$, or one of the strings `'left'` or `'right'`. The default value is `'left'`.

    Examples:
        >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 0, 1], [1, 1, 0]])
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1],
               [0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 1, 0, 0],
               [1, 0, 1, 0, 1, 0],
               [1, 1, 0, 0, 0, 1]])

        >>> code = komm.SystematicBlockCode(parity_submatrix=[[0, 1, 1], [1, 0, 1], [1, 1, 0]], information_set='right')
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[0, 1, 1, 1, 0, 0],
               [1, 0, 1, 0, 1, 0],
               [1, 1, 0, 0, 0, 1]])
        >>> code.check_matrix
        array([[1, 0, 0, 0, 1, 1],
               [0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 1, 0]])
    """

    _parity_submatrix: npt.ArrayLike = field(
        default=None, repr=False, alias="parity_submatrix"
    )
    _information_set: npt.ArrayLike | str = field(
        default="left", repr=False, alias="information_set"
    )

    def __repr__(self) -> str:
        args_str = f"parity_submatrix={self.parity_submatrix.tolist()}"
        args_str += f", information_set={self.information_set.tolist()}"
        return f"{self.__class__.__name__}({args_str})"

    @cached_property
    def parity_submatrix(self) -> npt.NDArray[np.int_]:
        return np.asarray(self._parity_submatrix)

    @cached_property
    def information_set(self) -> npt.NDArray[np.int_]:
        n, k, m = self.length, self.dimension, self.redundancy
        if self._information_set == "left":
            information_set = range(k)
        elif self._information_set == "right":
            information_set = range(m, n)
        else:
            information_set = self._information_set
        try:
            information_set = np.asarray(information_set)
        except TypeError:
            raise ValueError(
                "'information_set' must be either 'left', 'right', or an array of int"
            )
        if (
            information_set.size != k
            or information_set.min() < 0
            or information_set.max() > n
        ):
            raise ValueError("'information_set' must be a 'k'-sublist of 'range(n)'")
        return information_set

    @cached_property
    def parity_set(self) -> npt.NDArray[np.int_]:
        return np.setdiff1d(np.arange(self.length), self.information_set)

    @property
    def dimension(self) -> int:
        return self.parity_submatrix.shape[0]

    @property
    def redundancy(self) -> int:
        return self.parity_submatrix.shape[1]

    @property
    def length(self) -> int:
        return self.dimension + self.redundancy

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        k, n = self.dimension, self.length
        generator_matrix = np.empty((k, n), dtype=int)
        generator_matrix[:, self.information_set] = np.eye(k, dtype=int)
        generator_matrix[:, self.parity_set] = self.parity_submatrix
        return generator_matrix

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.int_]:
        m, n = self.redundancy, self.length
        check_matrix = np.empty((m, n), dtype=int)
        check_matrix[:, self.information_set] = self.parity_submatrix.T
        check_matrix[:, self.parity_set] = np.eye(m, dtype=int)
        return check_matrix

    def enc_mapping(self, u: ArrayIntLike) -> npt.NDArray[np.int_]:
        v = np.empty(self.length, dtype=int)
        v[self.information_set] = u
        v[self.parity_set] = np.dot(u, self.parity_submatrix) % 2
        return v

    def inv_enc_mapping(self, v: ArrayIntLike) -> npt.NDArray[np.int_]:
        v = np.asarray(v)
        if v.size != self.length:
            raise ValueError("length of 'v' must be equal to the code length")
        s = self.chk_mapping(v)
        if not np.all(s == 0):
            raise ValueError("input 'v' is not a valid codeword")
        u = np.take(v, self.information_set)
        return u



import itertools as it
from typing import cast

import numpy as np
import numpy.typing as npt
from attrs import field, frozen, validators
from typing_extensions import Self

from .._validation import is_pmf, validate_call
from .util import Word, is_prefix_free


@frozen
class VariableToFixedCode:
    r"""
    Variable-to-fixed length code. A *variable-to-fixed length code* with *target alphabet* $\mathcal{T}$, *source alphabet* $\mathcal{S}$, and *target block size* $n$ is defined by a (possibly partial) injective decoding mapping $\mathrm{Dec} : \mathcal{T}^n \to \mathcal{S}^+$, where the domain is the set of all $n$-tuples with entries in $\mathcal{T}$, and the co-domain is the set of all finite-length, non-empty tuples with entries in $\mathcal{S}$. Here, we assume that $\mathcal{T} = [0:T)$ and $\mathcal{S} = [0:S)$, for integers $T \geq 2$ and $S \geq 2$. The elements in the image of $\mathrm{Dec}$ are called *sourcewords*.

    Attributes:
        target_cardinality: The target cardinality $T$.
        source_cardinality: The source cardinality $S$.
        target_block_size: The target block size $n$.
        dec_mapping: The decoding mapping $\mathrm{Dec}$ of the code. Must be a dictionary of length at most $S^n$ whose keys are $n$-tuples of integers in $[0:T)$ and whose values are distinct non-empty tuples of integers in $[0:S)$.
    """

    target_cardinality: int = field(validator=validators.ge(2))
    source_cardinality: int = field(validator=validators.ge(2))
    target_block_size: int = field(validator=validators.ge(1))
    dec_mapping: dict[Word, Word] = field()

    def __attrs_post_init__(self) -> None:
        domain, codomain = self.dec_mapping.keys(), self.dec_mapping.values()
        T, S, n = (
            self.target_cardinality,
            self.source_cardinality,
            self.target_block_size,
        )
        if not set(domain) <= set(it.product(range(T), repeat=n)):
            raise ValueError(f"'dec_mapping': invalid domain")
        if not all(
            all(0 <= x < S for x in word) and len(word) > 0 for word in codomain
        ):
            raise ValueError(f"'dec_mapping': invalid co-domain")
        if len(set(codomain)) != len(codomain):
            raise ValueError(f"'dec_mapping': non-injective mapping")

    @classmethod
    def from_dec_mapping(cls, dec_mapping: dict[Word, Word]) -> Self:
        r"""
        Constructs a variable-to-fixed code from the decoding map $\Dec$.

        Parameters:
            dec_mapping: The decoding map $\Dec$. See the corresponding attribute for more details.

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({(0,0): (0,0,0), (0,1): (0,0,1), (1,0): (0,1), (1,1): (1,)})
            >>> (code.target_cardinality, code.source_cardinality, code.target_block_size)
            (2, 2, 2)
            >>> code.dec_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0): (0, 0, 0),
             (0, 1): (0, 0, 1),
             (1, 0): (0, 1),
             (1, 1): (1,)}

            >>> code = komm.VariableToFixedCode.from_dec_mapping({(0,0,0): (1,), (0,0,1): (2,), (0,1,0): (0,1), (0,1,1): (0,2), (1,0,0): (0,0,0), (1,0,1): (0,0,1), (1,1,0): (0,0,2)})
            >>> code.target_cardinality, code.source_cardinality, code.target_block_size
            (2, 3, 3)
            >>> code.dec_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0, 0): (1,),
             (0, 0, 1): (2,),
             (0, 1, 0): (0, 1),
             (0, 1, 1): (0, 2),
             (1, 0, 0): (0, 0, 0),
             (1, 0, 1): (0, 0, 1),
             (1, 1, 0): (0, 0, 2)}
        """
        domain, codomain = dec_mapping.keys(), dec_mapping.values()
        T = max(max(word) for word in domain) + 1
        S = max(max(word) for word in codomain) + 1
        n = len(next(iter(domain)))
        return cls(T, S, n, dec_mapping)

    @classmethod
    @validate_call(target_cardinality=field(validator=validators.ge(2)))
    def from_sourcewords(cls, target_cardinality: int, sourcewords: list[Word]) -> Self:
        r"""
        Constructs a variable-to-fixed code from the target cardinality $T$ and a list of sourcewords.

        Parameters:
            target_cardinality: The target cardinality $T$. Must be an integer greater than or equal to $2$.
            sourcewords: The sourcewords of the code. See the [corresponding property](./#sourcewords) for more details.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,0,1), (0,1), (1,)])
            >>> (code.target_cardinality, code.source_cardinality, code.target_block_size)
            (2, 2, 2)
            >>> code.dec_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0): (0, 0, 0),
             (0, 1): (0, 0, 1),
             (1, 0): (0, 1),
             (1, 1): (1,)}
        """
        T = target_cardinality
        S = max(max(word) for word in sourcewords) + 1
        n = next(n for n in it.count(1) if T**n >= len(sourcewords))
        dec_mapping = dict(zip(it.product(range(T), repeat=n), sourcewords))
        return cls(T, S, n, dec_mapping)

    @property
    def sourcewords(self) -> list[Word]:
        r"""
        The sourcewords of the code. It is a list of length at most $T^n$ containing tuples of integers in $[0:S)$. The tuple in position $i$ of `sourcewords` is equal to $\mathrm{Dec}(v)$, where $v$ is the $i$-th element in the lexicographic ordering of $[0:T)^n$.

        Examples:
            >>> code = komm.VariableToFixedCode.from_dec_mapping({(0,0): (0,0,0), (0,1): (0,0,1), (1,0): (0,1), (1,1): (1,)})
            >>> code.sourcewords
            [(0, 0, 0), (0, 0, 1), (0, 1), (1,)]
        """
        return list(self.dec_mapping.values())

    @property
    def inv_dec_mapping(self) -> dict[Word, Word]:
        r"""
        The inverse decoding mapping $\mathrm{Dec}^{-1}$ of the code. It is a dictionary of length at most $T^n$ whose keys are all the sourcewords of the code, and whose values are the corresponding target words.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,0,1), (0,1), (1,)])
            >>> code.inv_dec_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0, 0): (0, 0),
             (0, 0, 1): (0, 1),
             (0, 1): (1, 0),
             (1,): (1, 1)}
        """
        return {v: k for k, v in self.dec_mapping.items()}

    def is_unique_encodable(self) -> bool:
        r"""
        Returns whether the code is unique encodable or not. [Not implemented yet].
        """
        raise NotImplementedError

    def is_prefix_free(self) -> bool:
        r"""
        Returns whether the code is prefix-free or not. A code is *prefix-free* if no sourceword is a prefix of any other sourceword.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,0,1), (0,1), (1,)])
            >>> code.is_prefix_free()
            True

            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,1,0), (0,1), (1,)])
            >>> code.is_prefix_free()
            False
        """
        return is_prefix_free(self.sourcewords)

    @validate_call(pmf=field(converter=np.asarray, validator=is_pmf))
    def rate(self, pmf: npt.ArrayLike) -> float:
        r"""
        Computes the expected rate $R$ of the code, considering a given pmf. This quantity is given by
        $$
            R = \frac{n}{\bar{k}},
        $$
        where $n$ is the target block size, and $\bar{k}$ is the expected sourceword length, assuming iid source symbols drawn from $p_X$. It is measured in $T$-ary digits per source symbol.

        Parameters:
            pmf (Array1D[float]): The (first-order) probability mass function $p_X$ to be considered.

        Returns:
            rate: The expected rate $R$ of the code.

        Examples:
            >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,0,1), (0,1), (1,)])
            >>> code.rate([2/3, 1/3])
            np.float64(0.9473684210526315)
        """
        pmf = cast(npt.NDArray[np.float64], pmf)
        probabilities = [np.prod([pmf[x] for x in word]) for word in self.sourcewords]
        lengths = [len(word) for word in self.sourcewords]
        return self.target_block_size / np.dot(lengths, probabilities)



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from attrs import frozen

from .matrices import hamming_parity_submatrix
from .SystematicBlockCode import SystematicBlockCode


@frozen
class HammingCode(SystematicBlockCode):
    r"""
    Hamming code. For a given parameter $\mu \geq 2$, it is the [linear block code](/ref/BlockCode) with check matrix whose columns are all the $2^\mu - 1$ nonzero binary $\mu$-tuples. The Hamming code has the following parameters:

    - Length: $n = 2^\mu - 1$
    - Dimension: $k = 2^\mu - \mu - 1$
    - Redundancy: $m = \mu$
    - Minimum distance: $d = 3$

    In its extended version, the Hamming code has the following parameters:

    - Length: $n = 2^\mu$
    - Dimension: $k = 2^\mu - \mu - 1$
    - Redundancy: $m = \mu + 1$
    - Minimum distance: $d = 4$

    For more details, see <cite>LC04, Sec. 4.1</cite>.

    Notes:
        - For $\mu = 2$ it reduces to the [repetition code](/ref/RepetitionCode) of length $3$.
        - Its dual is the [simplex code](/ref/SimplexCode).
        - Hamming codes are perfect codes.

    Attributes:
        mu: The parameter $\mu$ of the code. Must satisfy $\mu \geq 2$.
        extended: Whether to use the extended version of the Hamming code. Default is `False`.

    This function returns the code in systematic form, with the information set on the left.

    Examples:
        >>> code = komm.HammingCode(3)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 4, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1, 0],
               [0, 1, 0, 0, 1, 0, 1],
               [0, 0, 1, 0, 0, 1, 1],
               [0, 0, 0, 1, 1, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 1, 0, 0],
               [1, 0, 1, 1, 0, 1, 0],
               [0, 1, 1, 1, 0, 0, 1]])
        >>> code.minimum_distance()
        3

        >>> code = komm.HammingCode(3, extended=True)
        >>> (code.length, code.dimension, code.redundancy)
        (8, 4, 4)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 0, 1, 1],
               [0, 0, 1, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[1, 1, 0, 1, 1, 0, 0, 0],
               [1, 0, 1, 1, 0, 1, 0, 0],
               [0, 1, 1, 1, 0, 0, 1, 0],
               [1, 1, 1, 0, 0, 0, 0, 1]])
        >>> code.minimum_distance()
        4
    """

    mu: int
    extended: bool = False

    @cached_property
    def parity_submatrix(self) -> npt.NDArray[np.int_]:
        return hamming_parity_submatrix(self.mu, self.extended)

    @cache
    def minimum_distance(self) -> int:
        return 4 if self.extended else 3



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt


class AbstractScalarQuantizer(ABC):
    @property
    @abstractmethod
    def levels(self) -> npt.NDArray[np.float64]:
        pass

    @property
    @abstractmethod
    def thresholds(self) -> npt.NDArray[np.float64]:
        pass

    @property
    @abstractmethod
    def num_levels(self) -> int:
        pass

    @abstractmethod
    def __call__(self, input_signal: npt.ArrayLike) -> npt.NDArray[np.float64]:
        pass



from attrs import field as attrs_field
from attrs import frozen
from typing_extensions import Self

from . import field
from .BinaryPolynomial import BinaryPolynomial


@frozen
class BinaryPolynomialFraction:
    r"""
    Binary polynomial fraction. A *binary polynomial fraction* is a ratio of two [binary polynomials](/ref/BinaryPolynomial).

    Attributes:
        numerator: The numerator of the fraction.
        denominator: The denominator of the fraction.

    Examples:
        >>> komm.BinaryPolynomialFraction(0b11010, 0b101)  # (X^4 + X^3 + X) / (X^2 + 1)
        BinaryPolynomialFraction(0b11010, 0b101)

    <h2>Algebraic structure</h2>

    The binary polynomial fractions form a *field*. The following operations are supported: addition (`+`), subtraction (`-`), multiplication (`*`), division (`/`), and exponentiation (`**`).
    """

    numerator: BinaryPolynomial = attrs_field(
        converter=BinaryPolynomial,
    )
    denominator: BinaryPolynomial = attrs_field(
        converter=BinaryPolynomial,
        default=BinaryPolynomial(0b1),
    )

    def __attrs_post_init__(self):
        if self.denominator == 0b0:
            raise ZeroDivisionError("denominator cannot be zero")
        gcd = BinaryPolynomial.gcd(self.numerator, self.denominator)
        object.__setattr__(self, "numerator", self.numerator // gcd)
        object.__setattr__(self, "denominator", self.denominator // gcd)

    @property
    def ambient(self):
        return BinaryPolynomialFractions()

    def __add__(self, other: Self) -> Self:
        n = self.numerator * other.denominator + other.numerator * self.denominator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __sub__(self, other: Self) -> Self:
        n = self.numerator * other.denominator - other.numerator * self.denominator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __neg__(self) -> Self:
        return self

    def __mul__(self, other: Self) -> Self:
        n = self.numerator * other.numerator
        d = self.denominator * other.denominator
        return self.__class__(n, d)

    def __rmul__(self, other: int) -> Self:
        return self.__class__(other * self.numerator, self.denominator)

    def __truediv__(self, other: Self) -> Self:
        n = self.numerator * other.denominator
        d = self.denominator * other.numerator
        return self.__class__(n, d)

    def __pow__(self, exponent: int) -> Self:
        return field.power(self, exponent)

    def inverse(self) -> Self:
        return self.__class__(self.denominator, self.numerator)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.numerator}, {self.denominator})"

    def __str__(self) -> str:
        return str(self.numerator) + "/" + str(self.denominator)


@frozen
class BinaryPolynomialFractions:
    def __call__(self, value: tuple[int, int]) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(*value)

    @property
    def zero(self) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(0, 1)

    @property
    def one(self) -> BinaryPolynomialFraction:
        return BinaryPolynomialFraction(1, 1)

    @property
    def characteristic(self) -> int:
        return 2



import itertools as it
from typing import cast

import numpy as np
import numpy.typing as npt
from attrs import field, frozen, validators
from typing_extensions import Self

from .._validation import is_pmf, validate_call
from .util import Word, is_prefix_free, is_uniquely_decodable


@frozen
class FixedToVariableCode:
    r"""
    Fixed-to-variable length code. A *fixed-to-variable length code* with *source alphabet* $\mathcal{S}$, *target alphabet* $\mathcal{T}$, and *source block size* $k$ is defined by an injective *encoding mapping* $\Enc : \mathcal{S}^k \to \mathcal{T}^+$, where the domain is the set of all $k$-tuples with entries in $\mathcal{S}$, and the co-domain is the set of all finite-length, non-empty tuples with entries in $\mathcal{T}$. Here we assume that $\mathcal{S} = [0:S)$ and $\mathcal{T} = [0:T)$, for integers $S \geq 2$ and $T \geq 2$. The elements in the image of $\Enc$ are called *codewords*.

    Attributes:
        source_cardinality: The source cardinality $S$.
        target_cardinality: The target cardinality $T$.
        source_block_size: The source block size $k$.
        enc_mapping: The encoding mapping $\Enc$ of the code. Must be a dictionary of length $S^k$ whose keys are $k$-tuples of integers in $[0:S)$ and whose values are distinct non-empty tuples of integers in $[0:T)$.
    """

    source_cardinality: int = field(validator=validators.ge(2))
    target_cardinality: int = field(validator=validators.ge(2))
    source_block_size: int = field(validator=validators.ge(1))
    enc_mapping: dict[Word, Word] = field()

    def __attrs_post_init__(self) -> None:
        domain, codomain = self.enc_mapping.keys(), self.enc_mapping.values()
        S, T, k = (
            self.source_cardinality,
            self.target_cardinality,
            self.source_block_size,
        )
        if set(domain) != set(it.product(range(S), repeat=k)):
            raise ValueError(f"'enc_mapping': invalid domain")
        if not all(
            all(0 <= x < T for x in word) and len(word) > 0 for word in codomain
        ):
            raise ValueError(f"'enc_mapping': invalid co-domain")
        if len(set(codomain)) != len(codomain):
            raise ValueError(f"'enc_mapping': non-injective mapping")

    @classmethod
    def from_enc_mapping(cls, enc_mapping: dict[Word, Word]) -> Self:
        r"""
        Constructs a fixed-to-variable length code from the encoding mapping $\Enc$.

        Parameters:
            enc_mapping: The encoding mapping $\Enc$. See the corresponding attribute for more details.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({(0,): (0,), (1,): (1,0), (2,): (1,1)})
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (3, 2, 1)
            >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0,): (0,),
             (1,): (1, 0),
             (2,): (1, 1)}

            >>> code = komm.FixedToVariableCode.from_enc_mapping({(0,0): (0,), (0,1): (1,0,0), (1,0): (1,1), (1,1): (1,0,1)})
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (2, 2, 2)
            >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0): (0,),
             (0, 1): (1, 0, 0),
             (1, 0): (1, 1),
             (1, 1): (1, 0, 1)}
        """
        domain, codomain = enc_mapping.keys(), enc_mapping.values()
        S = max(max(word) for word in domain) + 1
        T = max(max(word) for word in codomain) + 1
        k = len(next(iter(domain)))
        return cls(S, T, k, enc_mapping)

    @classmethod
    @validate_call(source_cardinality=field(validator=validators.ge(2)))
    def from_codewords(cls, source_cardinality: int, codewords: list[Word]) -> Self:
        r"""
        Constructs a fixed-to-variable length code from the source cardinality $S$ and a list of codewords.

        Parameters:
            source_cardinality: The source cardinality $S$. Must be an integer greater than or equal to $2$.
            codewords: The codewords of the code. See the [corresponding property](./#codewords) for more details.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
            >>> code.source_cardinality, code.target_cardinality, code.source_block_size
            (3, 2, 1)
            >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0,): (0,),
             (1,): (1, 0),
             (2,): (1, 1)}

            >>> code = komm.FixedToVariableCode.from_codewords(2, [(0,), (1,0,0), (1,1), (1,0,1)])
            >>> (code.source_cardinality, code.target_cardinality, code.source_block_size)
            (2, 2, 2)
            >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0, 0): (0,),
             (0, 1): (1, 0, 0),
             (1, 0): (1, 1),
             (1, 1): (1, 0, 1)}
        """
        S = source_cardinality
        T = max(max(codeword) for codeword in codewords) + 1
        k = next(k for k in it.count(1) if S**k >= len(codewords))
        enc_mapping = dict(zip(it.product(range(S), repeat=k), codewords))
        return cls(S, T, k, enc_mapping)

    @property
    def codewords(self) -> list[Word]:
        r"""
        The codewords of the code. It is a list of length $S^k$ containing tuples of integers in $[0:T)$. The tuple in position $i$ of `codewords` is equal to $\Enc(u)$, where $u$ is the $i$-th element in the lexicographic ordering of $[0:S)^k$.

        Examples:
            >>> code = komm.FixedToVariableCode.from_enc_mapping({(0,): (0,), (1,): (1,0), (2,): (1,1)})
            >>> code.codewords
            [(0,), (1, 0), (1, 1)]
        """
        return list(self.enc_mapping.values())

    @property
    def inv_enc_mapping(self) -> dict[Word, Word]:
        r"""
        The inverse encoding mapping $\Enc^{-1}$ of the code. It is a dictionary of length $S^k$ whose keys are all the codewords of the code and whose values are the corresponding source words.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
            >>> code.inv_enc_mapping  # doctest: +NORMALIZE_WHITESPACE
            {(0,): (0,),
             (1, 0): (1,),
             (1, 1): (2,)}
        """
        return {v: k for k, v in self.enc_mapping.items()}

    def is_uniquely_decodable(self) -> bool:
        r"""
        Returns whether the code is uniquely decodable or not. A code is *uniquely decodable* if
        $$
            s_1 \cdots s_n \neq s'_1 \cdots s'_m \implies \Enc(s_1) \cdots \Enc(s_n) \neq \Enc(s'_1) \cdots \Enc(s'_m).
        $$

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
            >>> code.is_uniquely_decodable()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,1)])
            >>> code.is_uniquely_decodable()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,0)])
            >>> code.is_uniquely_decodable()
            False
        """
        return is_uniquely_decodable(self.codewords)

    def is_prefix_free(self) -> bool:
        r"""
        Returns whether the code is prefix-free or not. A code is *prefix-free* if no codeword is a prefix of any other codeword.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
            >>> code.is_prefix_free()
            True

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,1)])
            >>> code.is_prefix_free()
            False

            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (0,1), (1,0)])
            >>> code.is_prefix_free()
            False
        """
        return is_prefix_free(self.codewords)

    @validate_call(pmf=field(converter=np.asarray, validator=is_pmf))
    def rate(self, pmf: npt.ArrayLike) -> float:
        r"""
        Computes the expected rate $R$ of the code, considering a given pmf. This quantity is given by
        $$
            R = \frac{\bar{n}}{k},
        $$
        where $\bar{n}$ is the expected codeword length, assuming iid source symbols drawn from $p_X$, and $k$ is the source block size. It is measured in $T$-ary digits per source symbol.

        Parameters:
            pmf: The (first-order) probability mass function $p_X$ to be considered.

        Returns:
            rate: The expected rate $R$ of the code.

        Examples:
            >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
            >>> code.rate([0.5, 0.25, 0.25])
            np.float64(1.5)
        """
        pmf = cast(npt.NDArray[np.float64], pmf)
        k = self.source_block_size
        probabilities = [np.prod(ps) for ps in it.product(pmf, repeat=k)]
        lengths = [len(word) for word in self.codewords]
        return np.dot(lengths, probabilities) / k



import numpy as np
import numpy.typing as npt

from ..._util import binlist2int
from ..BlockCode import BlockCode
from ..registry import RegistryBlockDecoder


def decode_syndrome_table(
    code: BlockCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    coset_leaders = code.coset_leaders()
    s = np.dot(code.check_matrix, r) % 2
    e_hat = coset_leaders[binlist2int(s)]
    v_hat = np.bitwise_xor(r, e_hat)
    return v_hat


RegistryBlockDecoder.register(
    "syndrome-table",
    {
        "description": "Syndrome table decoder",
        "decoder": decode_syndrome_table,
        "type_in": "hard",
        "type_out": "hard",
        "target": "codeword",
    },
)



from typing import Literal

import numpy.typing as npt

from .constellations import constellation_ask
from .labelings import get_labeling
from .Modulation import Modulation


class ASKModulation(Modulation):
    r"""
    Amplitude-shift keying (ASK) modulation. It is a complex [modulation scheme](/ref/Modulation) in which the points of the constellation symbols are *uniformly arranged* in a ray. More precisely, the $i$-th constellation symbol is given by
    $$
        x_i = iA \exp(\mathrm{j}\phi), \quad i \in [0 : M),
    $$
    where $M$ is the *order* (a power of $2$), $A$ is the *base amplitude*, and $\phi$ is the *phase offset* of the modulation. The ASK constellation is depicted below for $M = 4$.

    <figure markdown>
      ![4-ASK constellation.](/figures/ask_4.svg)
    </figure>

    Parameters:
        order: The order $M$ of the modulation. It must be a power of $2$.

        base_amplitude: The base amplitude $A$ of the constellation. The default value is `1.0`.

        phase_offset: The phase offset $\phi$ of the constellation. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural'` or `'reflected'`. The default value is `'reflected'`, corresponding to the Gray labeling.

    Examples:
        The ASK modulation with order $M = 4$, base amplitude $A = 1$, and Gray labeling is depicted below.

        <figure markdown>
          ![4-ASK modulation with Gray labeling.](/figures/ask_4_gray.svg)
        </figure>

        >>> ask = komm.ASKModulation(4)
        >>> ask.constellation
        array([0.+0.j, 1.+0.j, 2.+0.j, 3.+0.j])
        >>> ask.labeling
        array([[0, 0],
               [1, 0],
               [1, 1],
               [0, 1]])
        >>> ask.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([0.+0.j, 2.+0.j, 0.+0.j, 1.+0.j, 1.+0.j])
    """

    def __init__(
        self,
        order: int,
        base_amplitude: float = 1.0,
        phase_offset: float = 0.0,
        labeling: Literal["natural", "reflected"] | npt.ArrayLike = "reflected",
    ) -> None:
        super().__init__(
            constellation=constellation_ask(order, base_amplitude, phase_offset),
            labeling=get_labeling(labeling, ("natural", "reflected"), order),
        )
        self.base_amplitude = base_amplitude
        self.phase_offset = phase_offset
        self.labeling_parameter = labeling

    def __repr__(self) -> str:
        args = ", ".join([
            f"order={self.order}",
            f"base_amplitude={self.base_amplitude}",
            f"phase_offset={self.phase_offset}",
            f"labeling='{self.labeling_parameter}'",
        ])
        return f"{self.__class__.__name__}({args})"



from functools import reduce
from itertools import combinations

import numpy as np
import numpy.typing as npt

BinaryMatrix = npt.NDArray[np.int_]


def extended_parity_submatrix(parity_submatrix: BinaryMatrix) -> BinaryMatrix:
    last_column = (1 + np.sum(parity_submatrix, axis=1)) % 2
    extended_parity_submatrix = np.hstack([parity_submatrix, last_column[np.newaxis].T])
    return extended_parity_submatrix


def hamming_parity_submatrix(m: int, extended: bool = False) -> BinaryMatrix:
    parity_submatrix = np.zeros((2**m - m - 1, m), dtype=int)
    i = 0
    for w in range(2, m + 1):
        for idx in combinations(range(m), w):
            parity_submatrix[i, list(idx)] = 1
            i += 1
    if extended:
        parity_submatrix = extended_parity_submatrix(parity_submatrix)
    return parity_submatrix


def golay_parity_submatrix(extended: bool = False) -> BinaryMatrix:
    parity_submatrix = np.array([
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
        [0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1],
        [1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1],
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1],
        [1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1],
        [0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1],
        [0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 1],
        [0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1],
        [1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1],
        [1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1],
        [1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1],
    ])
    if extended:
        parity_submatrix = extended_parity_submatrix(parity_submatrix)
    return parity_submatrix


def reed_muller_generator_matrix(rho: int, mu: int) -> BinaryMatrix:
    # See [LC04, p. 105–114]. Assumes 0 <= rho < mu.
    v = np.empty((mu, 2**mu), dtype=int)
    for i in range(mu):
        block = np.hstack((
            np.zeros(2 ** (mu - i - 1), dtype=int),
            np.ones(2 ** (mu - i - 1), dtype=int),
        ))
        v[mu - i - 1] = np.tile(block, 2**i)

    G_list: list[npt.NDArray[np.int_]] = []
    for ell in range(rho, 0, -1):
        for indices in combinations(range(mu), ell):
            row = reduce(np.multiply, v[indices, :])
            G_list.append(row)
    G_list.append(np.ones(2**mu, dtype=int))

    return np.array(G_list, dtype=int)


def cordaro_wagner_generator_matrix(n: int) -> BinaryMatrix:
    # See [CW67].
    d = int(np.ceil(2 * n / 3)) - 1
    q = (n + 1) // 3
    return np.hstack((
        np.repeat([[1], [0]], repeats=d - q, axis=1),
        np.repeat([[0], [1]], repeats=q, axis=1),
        np.repeat([[1], [1]], repeats=n - d, axis=1),
    ))



import numpy as np
import numpy.typing as npt

from .._finite_state_machine.FiniteStateMachine import FiniteStateMachine
from .ConvolutionalCode import ConvolutionalCode


class PuncturedConvolutionalCode:
    r"""Punctured convolutional code. It is obtained by periodically deleting some of the bits from the output of a $(n, k)$ *mother [convolutional code](/ref/ConvolutionalCode)*. The *puncturing pattern* $\mathbf{\Pi}$ is an $n \times T$ binary matrix where each entry indicates whether a bit is kept ($1$) or deleted ($0$). The parameter $T$ is called the *puncturing period*. The resulting code has rate $R = (k T) / \wH({\mathbf{\Pi}})$, where $\wH({\mathbf{\Pi}})$ is the Hamming weight of the puncturing pattern. For more details, see <cite>LC04, Sec. 12.7</cite>.

    Parameters:
        mother_code: The mother convolutional code.
        puncture_matrix: The puncturing pattern $\mathbf{\Pi}$.

    Examples:
        Consider a mother convolutional code with parameters $(n, k) = (2, 1)$ and transfer function matrix $\mathbf{G}(D) = \begin{bmatrix}D^2 + 1 & D^2 + D + 1\end{bmatrix}$.

        If the puncturing pattern is given by
        $$
            \mathbf{\Pi} = \begin{bmatrix}
                1 & 0 \\\\
                1 & 1 \\\\
            \end{bmatrix},
        $$
        then the punctured convolutional code has rate $R = 2/3$.

        >>> mother_code = ConvolutionalCode(feedforward_polynomials=[[0b101, 0b111]])
        >>> puncture_matrix = np.array([[1, 0], [1, 1]])
        >>> punctured_code = PuncturedConvolutionalCode(mother_code, puncture_matrix)
        >>> punctured_code.puncturing_period
        2
        >>> punctured_code.rate
        0.6666666666666666

        If the puncturing pattern is given by
        $$
            \mathbf{\Pi} = \begin{bmatrix}
                1 & 0 & 1 \\\\
                1 & 1 & 0 \\\\
            \end{bmatrix},
        $$
        then the punctured convolutional code has rate $R = 3/4$.

        >>> mother_code = komm.ConvolutionalCode(feedforward_polynomials=[[0b101, 0b111]])
        >>> puncture_matrix = np.array([[1, 0, 1], [1, 1, 0]])
        >>> punctured_code = komm.PuncturedConvolutionalCode(mother_code, puncture_matrix)
        >>> punctured_code.puncturing_period
        3
        >>> punctured_code.rate
        0.75
    """

    def __init__(self, mother_code: ConvolutionalCode, puncture_matrix: npt.ArrayLike):
        self.mother_code = mother_code
        self.puncture_matrix = np.asarray(puncture_matrix)

    def __repr__(self) -> str:
        args = (
            f"(mother_code={self.mother_code}, "
            f"puncture_matrix={self.puncture_matrix.tolist()})"
        )
        return f"{self.__class__.__name__}{args}"

    @property
    def puncturing_period(self) -> int:
        r"""
        The puncturing period $T$.
        """
        return self.puncture_matrix.shape[1]

    @property
    def rate(self) -> float:
        r"""
        The rate $R$ of the punctured convolutional code.
        """
        k = self.mother_code.num_input_bits
        return k * self.puncturing_period / np.count_nonzero(self.puncture_matrix)

    @property
    def finite_state_machine(self) -> FiniteStateMachine:
        r"""
        The finite state machine of the punctured convolutional code.
        """
        return self.mother_code.finite_state_machine()



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt
from attrs import field, frozen
from typing_extensions import override

from .._error_control_block.BlockCode import BlockCode
from .._types import ArrayIntLike
from .._util.bit_operations import int2binlist, unpack
from .ConvolutionalCode import ConvolutionalCode
from .terminations import (
    DirectTruncation,
    TailBiting,
    TerminationStrategy,
    ZeroTermination,
)

TerminationMode = Literal["direct-truncation", "zero-termination", "tail-biting"]


@frozen
class TerminatedConvolutionalCode(BlockCode):
    r"""
    Terminated convolutional code. It is a [linear block code](/ref/BlockCode) obtained by terminating a $(n_0, k_0)$ [convolutional code](/ref/ConvolutionalCode). A total of $h$ information blocks (each containing $k_0$ information bits) is encoded. The dimension of the resulting block code is thus $k = h k_0$; its length depends on the termination mode employed. There are three possible termination modes:

    - **Direct truncation**. The encoder always starts at state $0$, and its output ends immediately after the last information block. The encoder may not necessarily end in state $0$. The resulting block code will have length $n = h n_0$.

    - **Zero termination**. The encoder always starts and ends at state $0$. To achieve this, a sequence of $k \mu$ tail bits is appended to the information bits, where $\mu$ is the memory order of the convolutional code. The resulting block code will have length $n = (h + \mu) n_0$.

    - **Tail-biting**. The encoder always starts and ends at the same state. To achieve this, the initial state of the encoder is chosen as a function of the information bits. The resulting block code will have length $n = h n_0$.

    For more details, see <cite>LC04, Sec. 12.7</cite> and <cite>WBR01</cite>.

    Attributes:
        convolutional_code: The convolutional code to be terminated.

        num_blocks: The number $h$ of information blocks.

        mode: The termination mode. It must be one of `'direct-truncation'` | `'zero-termination'` | `'tail-biting'`. The default value is `'zero-termination'`.

    Examples:
        >>> convolutional_code = komm.ConvolutionalCode([[0b1, 0b11]])

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='direct-truncation')
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 0, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        2

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='zero-termination')
        >>> (code.length, code.dimension, code.redundancy)
        (8, 3, 5)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0, 0, 0],
               [0, 0, 1, 1, 0, 1, 0, 0],
               [0, 0, 0, 0, 1, 1, 0, 1]])
        >>> code.minimum_distance()
        3

        >>> code = komm.TerminatedConvolutionalCode(convolutional_code, num_blocks=3, mode='tail-biting')
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 1, 0, 1, 0, 0],
               [0, 0, 1, 1, 0, 1],
               [0, 1, 0, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    convolutional_code: ConvolutionalCode
    num_blocks: int
    mode: TerminationMode = field(default="zero-termination")

    def __attrs_post_init__(self):
        if self.mode not in TerminationMode.__args__:
            raise ValueError(
                f"mode '{self.mode}' is unknown\n"
                f"supported termination modes: {set(TerminationMode.__args__)}"
            )

    @cached_property
    def _strategy(self) -> TerminationStrategy:
        return {
            "direct-truncation": DirectTruncation,
            "zero-termination": ZeroTermination,
            "tail-biting": TailBiting,
        }[self.mode](self.convolutional_code, self.num_blocks)

    @property
    @override
    def length(self) -> int:
        return self._strategy.codeword_length()

    @property
    @override
    def dimension(self) -> int:
        return self.num_blocks * self.convolutional_code.num_input_bits

    @cached_property
    @override
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        return self._strategy.generator_matrix(self)

    @override
    def enc_mapping(self, u: ArrayIntLike) -> npt.NDArray[np.int_]:
        n0 = self.convolutional_code.num_output_bits
        fsm = self.convolutional_code.finite_state_machine()
        input_bits = self._strategy.pre_process_input(u)
        initial_state = self._strategy.initial_state(input_bits)
        output_sequence, _ = fsm.process(input_bits, initial_state)
        v = unpack(output_sequence, width=n0)
        return v

    @property
    @override
    def default_decoder(self) -> str:
        return "viterbi-hard"

    @classmethod
    @override
    def supported_decoders(cls) -> list[str]:
        return cls.__base__.supported_decoders() + ["viterbi-hard", "viterbi-soft", "bcjr"]  # type: ignore

    @cached_property
    def cache_bit(self) -> npt.NDArray[np.int_]:
        n0 = self.convolutional_code.num_output_bits
        return np.array([int2binlist(y, width=n0) for y in range(2**n0)])

    @cached_property
    def cache_polar(self) -> npt.NDArray[np.int_]:
        return (-1) ** self.cache_bit



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from attrs import field, frozen
from tqdm import tqdm

from .._types import ArrayIntLike
from .._util.matrices import null_matrix, pseudo_inverse, rref
from .AbstractBlockCode import AbstractBlockCode
from .SlepianArray import SlepianArray


@frozen(kw_only=True, eq=False)
class BlockCode(AbstractBlockCode):
    r"""
    General binary linear block code. It is characterized by its *generator matrix* $G \in \mathbb{B}^{k \times n}$, and by its *check matrix* $H \in \mathbb{B}^{m \times n}$, which are related by $G H^\transpose = 0$. The parameters $n$, $k$, and $m$ are called the code *length*, *dimension*, and *redundancy*, respectively, and are related by $k + m = n$. For more details, see <cite>LC04, Ch. 3</cite>.

    The constructor expects either the generator matrix or the check matrix.

    Parameters:
        generator_matrix (Array2D[int]): The generator matrix $G$ of the code, which is a $k \times n$ binary matrix.
        check_matrix (Array2D[int]): The check matrix $H$ of the code, which is a $m \times n$ binary matrix.

    Examples:
        >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1],
               [0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 1, 0, 0],
               [1, 0, 1, 0, 1, 0],
               [1, 1, 0, 0, 0, 1]])

        >>> code = komm.BlockCode(check_matrix=[[0, 1, 1, 1, 0, 0], [1, 0, 1, 0, 1, 0], [1, 1, 0, 0, 0, 1]])
        >>> (code.length, code.dimension, code.redundancy)
        (6, 3, 3)
        >>> code.generator_matrix
        array([[1, 0, 0, 0, 1, 1],
               [0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 1, 0]])
        >>> code.check_matrix
        array([[0, 1, 1, 1, 0, 0],
               [1, 0, 1, 0, 1, 0],
               [1, 1, 0, 0, 0, 1]])
    """

    _generator_matrix: npt.ArrayLike | None = field(
        default=None, repr=False, alias="generator_matrix"
    )
    _check_matrix: npt.ArrayLike | None = field(
        default=None, repr=False, alias="check_matrix"
    )

    def __repr__(self) -> str:
        if self._generator_matrix is not None:
            args_str = f"generator_matrix={self.generator_matrix.tolist()}"
        else:  # self._check_matrix is not None
            args_str = f"check_matrix={self.check_matrix.tolist()}"
        return f"{self.__class__.__name__}({args_str})"

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        if self._generator_matrix is not None:
            return np.asarray(self._generator_matrix)
        return rref(null_matrix(self.check_matrix))

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.int_]:
        if self._check_matrix is not None:
            return np.asarray(self._check_matrix)
        return null_matrix(self.generator_matrix)

    @property
    def length(self) -> int:
        r"""
        The length $n$ of the code.
        """
        if self._generator_matrix is not None:
            return self.generator_matrix.shape[1]
        if self._check_matrix is not None:
            return self.check_matrix.shape[1]
        return self.dimension + self.redundancy

    @property
    def dimension(self) -> int:
        r"""
        The dimension $k$ of the code.
        """
        try:
            return self.generator_matrix.shape[0]
        except AttributeError:
            return self.length - self.redundancy

    @property
    def redundancy(self) -> int:
        r"""
        The redundancy $m$ of the code.
        """
        try:
            return self.check_matrix.shape[0]
        except AttributeError:
            return self.length - self.dimension

    @cached_property
    def _generator_matrix_pseudo_inverse(self) -> npt.NDArray[np.int_]:
        return pseudo_inverse(self.generator_matrix)

    @property
    def rate(self) -> float:
        r"""
        The rate $R = k/n$ of the code.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.rate
            0.5
        """
        return self.dimension / self.length

    def enc_mapping(self, u: ArrayIntLike) -> npt.NDArray[np.int_]:
        r"""
        The encoding mapping $\Enc : \mathbb{B}^k \to \mathbb{B}^n$ of the code. This is a function that takes a message $u \in \mathbb{B}^k$ and returns the corresponding codeword $v \in \mathbb{B}^n$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.enc_mapping([1, 0, 1])
            array([1, 0, 1, 1, 0, 1])
        """
        v = np.dot(u, self.generator_matrix) % 2
        return v

    def inv_enc_mapping(self, v: ArrayIntLike) -> npt.NDArray[np.int_]:
        r"""
        The inverse encoding mapping $\Enc^{-1} : \mathbb{B}^n \to \mathbb{B}^k$ of the code. This is a function that takes a codeword $v \in \mathbb{B}^n$ and returns the corresponding message $u \in \mathbb{B}^k$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.inv_enc_mapping([1, 0, 1, 1, 0, 1])
            array([1, 0, 1])
        """
        v = np.asarray(v)
        if v.size != self.length:
            raise ValueError("length of 'v' must be equal to the code length")
        s = self.chk_mapping(v)
        if not np.all(s == 0):
            raise ValueError("input 'v' is not a valid codeword")
        u = np.dot(v, self._generator_matrix_pseudo_inverse) % 2
        return u

    def chk_mapping(self, r: ArrayIntLike) -> npt.NDArray[np.int_]:
        r"""
        The check mapping $\mathrm{Chk}: \mathbb{B}^n \to \mathbb{B}^m$ of the code. This is a function that takes a received word $r \in \mathbb{B}^n$ and returns the corresponding syndrome $s \in \mathbb{B}^m$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.chk_mapping([1, 0, 1, 1, 0, 1])
            array([0, 0, 0])
        """
        r = np.asarray(r)
        if r.size != self.length:
            raise ValueError("length of 'r' must be equal to the code length")
        s = np.dot(self.check_matrix, r) % 2
        return s

    @cache
    def codewords(self, _batch_size: int = 1024) -> npt.NDArray[np.int_]:
        r"""
        Returns the codewords of the code. This is a $2^k \times n$ matrix whose rows are all the codewords. The codeword in row $i$ corresponds to the message obtained by expressing $i$ in binary with $k$ bits (MSB in the right).

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.codewords()
            array([[0, 0, 0, 0, 0, 0],
                   [1, 0, 0, 0, 1, 1],
                   [0, 1, 0, 1, 0, 1],
                   [1, 1, 0, 1, 1, 0],
                   [0, 0, 1, 1, 1, 0],
                   [1, 0, 1, 1, 0, 1],
                   [0, 1, 1, 0, 1, 1],
                   [1, 1, 1, 0, 0, 0]])
        """
        k, n = self.dimension, self.length
        codewords = np.empty((2**k, n), dtype=int)
        for i in tqdm(
            range(0, 2**k, _batch_size), desc="Generating codewords", delay=2.5
        ):
            batch_end = min(i + _batch_size, 2**k)
            js = np.arange(i, batch_end, dtype=np.uint64).reshape(-1, 1).view(np.uint8)
            messages_batch = np.unpackbits(js, axis=1, count=k, bitorder="little")
            codewords[i:batch_end] = np.dot(messages_batch, self.generator_matrix) % 2
        return codewords

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.int_]:
        r"""
        Returns the codeword weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of codewords of Hamming weight $w$, for $w \in [0 : n]$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.codeword_weight_distribution()
            array([1, 0, 0, 4, 3, 0, 0])
        """
        return np.bincount(np.sum(self.codewords(), axis=1), minlength=self.length + 1)

    @cache
    def minimum_distance(self) -> int:
        r"""
        Returns the minimum distance $d$ of the code. This is equal to the minimum Hamming weight of the non-zero codewords.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.minimum_distance()
            3
        """
        return int(np.flatnonzero(self.codeword_weight_distribution())[1])

    @cache
    def coset_leaders(self) -> npt.NDArray[np.int_]:
        r"""
        Returns the coset leaders of the code. This is a $2^m \times n$ matrix whose rows are all the coset leaders. The coset leader in row $i$ corresponds to the syndrome obtained by expressing $i$ in binary with $m$ bits (MSB in the right), and whose Hamming weight is minimal. This may be used as a LUT for syndrome-based decoding.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.coset_leaders()
            array([[0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 1, 0, 0],
                   [0, 0, 0, 0, 1, 0],
                   [0, 0, 1, 0, 0, 0],
                   [0, 0, 0, 0, 0, 1],
                   [0, 1, 0, 0, 0, 0],
                   [1, 0, 0, 0, 0, 0],
                   [1, 0, 0, 1, 0, 0]])
        """
        sa = SlepianArray(self)
        return sa.col(0)

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.int_]:
        r"""
        Returns the coset leader weight distribution of the code. This is an array of shape $(n + 1)$ in which element in position $w$ is equal to the number of coset leaders of weight $w$, for $w \in [0 : n]$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.coset_leader_weight_distribution()
            array([1, 6, 1, 0, 0, 0, 0])
        """
        return np.bincount(
            np.sum(self.coset_leaders(), axis=1), minlength=self.length + 1
        )

    @cache
    def packing_radius(self) -> int:
        r"""
        Returns the packing radius of the code. This is also called the *error-correcting capability* of the code, and is equal to $\lfloor (d - 1) / 2 \rfloor$.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.packing_radius()
            1
        """
        return (self.minimum_distance() - 1) // 2

    @cache
    def covering_radius(self) -> int:
        r"""
        Returns the covering radius of the code. This is equal to the maximum Hamming weight of the coset leaders.

        Examples:
            >>> code = komm.BlockCode(generator_matrix=[[1, 0, 0, 0, 1, 1], [0, 1, 0, 1, 0, 1], [0, 0, 1, 1, 1, 0]])
            >>> code.covering_radius()
            2
        """
        return int(np.flatnonzero(self.coset_leader_weight_distribution())[-1])

    @property
    def default_decoder(self) -> str:
        return (
            "exhaustive-search-hard"
            if self.dimension <= self.redundancy
            else "syndrome-table"
        )

    @classmethod
    def supported_decoders(cls) -> list[str]:
        return ["exhaustive-search-hard", "exhaustive-search-soft", "syndrome-table"]



from typing import Any

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._util.information_theory import LogBase, binary_entropy
from .._validation import is_pmf, is_probability, validate_call
from .AbstractDiscreteMemorylessChannel import AbstractDiscreteMemorylessChannel


@frozen
class ZChannel(AbstractDiscreteMemorylessChannel):
    r"""
    Z-channel. It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input and output alphabets $\mathcal{X} = \mathcal{Y} = \\{ 0, 1 \\}$. The channel is characterized by a parameter $p$, called the *decay probability*. Bit $0$ is always received correctly, but bit $1$ turns into $0$ with probability $p$. Equivalently, the channel can be modeled as
    $$
        Y_n = A_n X_n,
    $$
    where $A_n$ are iid Bernoulli random variables with $\Pr[A_n = 0] = p$.

    Attributes:
        decay_probability (Optional[float]): The channel decay probability $p$. Must satisfy $0 \leq p \leq 1$. The default value is `0.0`, which corresponds to a noiseless channel.

    Parameters: Input:
        input_sequence (Array1D[int]): The input sequence.

    Parameters: Output:
        output_sequence (Array1D[int]): The output sequence.

    Examples:
        >>> np.random.seed(1)
        >>> zc = komm.ZChannel(0.1)
        >>> zc([0, 1, 1, 1, 0, 0, 0, 0, 0, 1])
        array([0, 1, 0, 1, 0, 0, 0, 0, 0, 1])
    """

    decay_probability: float = field(default=0.0, validator=is_probability)

    @property
    def input_cardinality(self) -> int:
        return 2

    @property
    def output_cardinality(self) -> int:
        return 2

    @property
    def transition_matrix(self) -> npt.NDArray[np.float64]:
        r"""
        The transition probability matrix of the channel. It is given by
        $$
            p_{Y \mid X} = \begin{bmatrix} 1 & 0 \\\\ p & 1-p \end{bmatrix}.
        $$

        Examples:
            >>> zc = komm.ZChannel(0.1)
            >>> zc.transition_matrix
            array([[1. , 0. ],
                   [0.1, 0.9]])
        """
        p = self.decay_probability
        return np.array([[1, 0], [p, 1 - p]])

    @validate_call(input_pmf=field(converter=np.asarray, validator=is_pmf))
    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        Returns the mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$ of the channel. It is given by
        $$
            \mathrm{I}(X ; Y) = \Hb \( \pi (1-p) \) - \pi \Hb(p),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        **Parameters:**

        Same as the [corresponding method](/ref/DiscreteMemorylessChannel/#mutual_information) of the general class.

        Examples:
            >>> zc = komm.ZChannel(0.1)
            >>> zc.mutual_information([0.5, 0.5])  # doctest: +NUMBER
            np.float64(0.7582766571931676)
        """
        input_pmf = np.array(input_pmf)
        p = self.decay_probability
        pi = input_pmf[1]
        return (binary_entropy(pi * (1 - p)) - pi * binary_entropy(p)) / np.log2(base)

    def capacity(self, base: LogBase = 2.0, **kwargs: Any) -> float:
        r"""
        Returns the channel capacity $C$. It is given by
        $$
            C = \log_2 \( 1 + (1-p) p^{p / (1-p)} \),
        $$
        in bits.

        Examples:
            >>> zc = komm.ZChannel(0.1)
            >>> zc.capacity()  # doctest: +NUMBER
            np.float64(0.7628482520105094)
        """
        p = self.decay_probability
        if p == 1.0:
            return 0.0
        q = 1 - p
        return np.log2(1 + q * p ** (p / q)) / np.log2(base)

    def __call__(self, input_sequence: npt.ArrayLike):
        p = self.decay_probability
        input_sequence = np.array(input_sequence)
        keep_pattern = (np.random.rand(np.size(input_sequence)) > p).astype(int)
        return input_sequence * keep_pattern



from .FixedToVariableCode import FixedToVariableCode
from .FixedToVariableDecoder import FixedToVariableDecoder
from .FixedToVariableEncoder import FixedToVariableEncoder
from .HuffmanCode import HuffmanCode
from .TunstallCode import TunstallCode
from .VariableToFixedCode import VariableToFixedCode
from .VariableToFixedDecoder import VariableToFixedDecoder
from .VariableToFixedEncoder import VariableToFixedEncoder

__all__ = [
    "FixedToVariableCode",
    "FixedToVariableDecoder",
    "FixedToVariableEncoder",
    "HuffmanCode",
    "TunstallCode",
    "VariableToFixedCode",
    "VariableToFixedDecoder",
    "VariableToFixedEncoder",
]



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from attrs import frozen

from .BlockCode import BlockCode
from .matrices import cordaro_wagner_generator_matrix


@frozen
class CordaroWagnerCode(BlockCode):
    r"""
    Cordaro–Wagner code. For a given length $n \geq 2$, it is the [linear block code](/ref/BlockCode) with dimension $k = 2$ which is optimum for the [BSC](/ref/BinarySymmetricChannel) with sufficiently small crossover probability. For more details, see <cite>CW67</cite>.

    - Length: $n$
    - Dimension: $k = 2$
    - Redundancy: $m = n - 2$
    - Minimum distance: $d = \left\lceil 2n / 3 \right\rceil - 1$

    Attributes:
        n: The length $n$ of the code. Must satisfy $n \geq 2$.

    Examples:
        >>> code = komm.CordaroWagnerCode(11)
        >>> (code.length, code.dimension, code.redundancy)
        (11, 2, 9)
        >>> code.generator_matrix
        array([[1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
               [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        7
        >>> code.codeword_weight_distribution()
        array([1, 0, 0, 0, 0, 0, 0, 2, 1, 0, 0, 0])
        >>> code.coset_leader_weight_distribution()
        array([  1,  11,  55, 165, 226,  54,   0,   0,   0,   0,   0,   0])
    """

    n: int

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        return cordaro_wagner_generator_matrix(self.n)

    @cache
    def minimum_distance(self) -> int:
        return int(np.ceil(2 * self.n / 3)) - 1

    @property
    def default_decoder(self) -> str:
        return "exhaustive-search-hard"



from typing import Callable, Tuple

import numpy as np
import numpy.typing as npt

from .AbstractScalarQuantizer import AbstractScalarQuantizer


def mean_squared_quantization_error(
    quantizer: AbstractScalarQuantizer,
    input_pdf: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]],
    input_range: Tuple[float, float],
    points_per_interval: int,
) -> np.float64:
    # See [Say06, eq. (9.3)].
    x_min, x_max = input_range
    thresholds = np.concatenate(([x_min], quantizer.thresholds, [x_max]))
    mse = np.float64(0.0)
    for i, level in enumerate(quantizer.levels):
        left, right = thresholds[i], thresholds[i + 1]
        x = np.linspace(left, right, num=points_per_interval, dtype=np.float64)
        pdf = input_pdf(x)
        integrand: npt.NDArray[np.float64] = (x - level) ** 2 * pdf
        integral = np.float64(np.trapezoid(integrand, x))
        mse += integral
    return mse


def lloyd_max_quantizer(
    input_pdf: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]],
    num_levels: int,
    input_range: Tuple[float, float],
    points_per_interval: int,
    max_iter: int,
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    # See [Say06, eqs. (9.27) and (9.28)].
    x_min, x_max = input_range
    delta = (x_max - x_min) / num_levels

    # Initial guess
    levels = np.linspace(x_min + delta / 2, x_max - delta / 2, num=num_levels)
    thresholds = np.empty(num_levels + 1, dtype=float)
    new_levels = np.empty_like(levels)

    for _ in range(max_iter):
        thresholds[0] = x_min
        thresholds[1:-1] = 0.5 * (levels[:-1] + levels[1:])
        thresholds[-1] = x_max

        for i in range(num_levels):
            left, right = thresholds[i], thresholds[i + 1]
            x = np.linspace(left, right, num=points_per_interval, dtype=np.float64)
            pdf = input_pdf(x)
            numerator = np.trapezoid(x * pdf, x)
            denominator = np.trapezoid(pdf, x)
            if denominator != 0:
                new_levels[i] = numerator / denominator
            else:  # Keep old level
                new_levels[i] = levels[i]
        if np.allclose(levels, new_levels):
            break
        levels = new_levels.copy()

    return new_levels, thresholds[1:-1]



from typing import Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt

DType = TypeVar("DType", bound=Union[np.float64, np.complex128])

from .._types import ArrayIntLike


def acorr(
    seq: npt.ArrayLike,
    shifts: Optional[ArrayIntLike] = None,
    normalized: bool = False,
) -> npt.NDArray[DType]:
    r"""
    Computes the autocorrelation $R[\ell]$ of a real or complex sequence $x[n]$. This is defined as
    $$
        R[\ell] = \sum_{n \in \mathbb{Z}} x[n] x^*_\ell[n],
    $$
    where $x^\*\_\ell[n] = x^\*[n - \ell]$ is the complex conjugate of $x[n]$ shifted by $\ell$ positions. The autocorrelation $R[\ell]$ is even symmetric and satisfies $R[\ell] = 0$ for $|\ell| \geq L$, where $L$ is the length of the sequence.

    Parameters:
        seq (Array1D[float | complex]): A 1D-array containing the sequence $x[n]$, of length $L$.

        shifts (Optional[Array1D[int]]): A 1D-array containing the values of $\ell$ for which the autocorrelation will be computed. The default value is `range(len(seq))`, that is, $[0 : L)$.

        normalized (Optional[bool]): If `True`, returns the autocorrelation divided by the sequence energy, so that $R[0] = 1$. The default value is `False`.

    Returns:
        acorr (SameAsInput): The autocorrelation $R[\ell]$ of the sequence.

    Examples:
        >>> komm.acorr([1.0, 2.0, 3.0, 4.0], shifts=[-2, -1, 0, 1, 2])
        array([11., 20., 30., 20., 11.])
    """
    seq = np.asarray(seq)
    seq_conj = np.conj(seq)
    shifts = np.arange(seq.size) if shifts is None else shifts
    acorr = np.empty_like(shifts, dtype=seq.dtype)
    for i, ell in enumerate(shifts):
        if ell < 0:
            acorr[i] = np.dot(seq[:ell], seq_conj[-ell:])
        elif ell > 0:
            acorr[i] = np.dot(seq[ell:], seq_conj[:-ell])
        else:
            acorr[i] = np.dot(seq, seq_conj)
    energy = np.dot(seq, seq_conj)
    return acorr / energy if normalized else acorr


def cyclic_acorr(
    seq: npt.ArrayLike,
    shifts: Optional[ArrayIntLike] = None,
    normalized: bool = False,
) -> npt.NDArray[DType]:
    r"""
    Computes the cyclic autocorrelation $\tilde{R}[\ell]$ of a real or complex sequence $x[n]$. This is defined as
    $$
        \tilde{R}[\ell] = \sum_{n \in [0:L)} x[n] \tilde{x}^*_\ell[n],
    $$
    where $\tilde{x}^\*\_\ell[n]$ is the complex conjugate of $x[n]$ cyclic-shifted by $\ell$ positions, and $L$ is the period of the sequence. The cyclic autocorrelation $\tilde{R}[\ell]$ is even symmetric and periodic with period $L$.

    Parameters:
        seq (Array1D[float | complex]): A 1D-array containing the sequence $x[n]$, of length $L$.

        shifts (Optional[Array1D[int]]): A 1D-array containing the values of $\ell$ for which the cyclic autocorrelation will be computed. The default value is `range(len(seq))`, that is, $[0 : L)$.

        normalized (Optional[bool]): If `True`, returns the cyclic autocorrelation divided by the sequence energy, so that $R[0] = 1$. The default value is `False`.

    Returns:
        cyclic_acorr (SameAsInput): The cyclic autocorrelation $\tilde{R}[\ell]$ of the sequence.

    Examples:
        >>> komm.cyclic_acorr([1.0, 2.0, 3.0, 4.0], shifts=[-2, -1, 0, 1, 2])
        array([22., 24., 30., 24., 22.])
    """
    seq = np.asarray(seq)
    shifts = np.arange(seq.size) if shifts is None else shifts
    conj_seq = np.conj(seq)
    acorr = np.array([np.dot(seq, np.roll(seq, ell).conj()) for ell in shifts])
    energy = np.dot(seq, conj_seq)
    return acorr / energy if normalized else acorr



from typing import Literal

import numpy as np
import numpy.typing as npt
from typing_extensions import override

from .._util.special_functions import logcosh
from .constellations import constellation_pam
from .labelings import get_labeling
from .Modulation import Modulation


class PAModulation(Modulation):
    r"""
    Pulse-amplitude modulation (PAM). It is a real [modulation scheme](/ref/Modulation) in which the constellation symbols are *uniformly arranged* in the real line and have zero mean. More precisely, the the $i$-th constellation symbol is given by
    $$
        x_i = A \left( 2i - M + 1 \right), \quad i \in [0 : M),
    $$
    where $M$ is the *order* (a power of $2$), and $A$ is the *base amplitude* of the modulation. The PAM constellation is depicted below for $M = 8$.

    <figure markdown>
      ![8-PAM constellation.](/figures/pam_8.svg)
    </figure>

    Parameters:
        order: The order $M$ of the modulation. It must be a power of $2$.

        base_amplitude: The base amplitude $A$ of the constellation. The default value is `1.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural'` or `'reflected'`. The default value is `'reflected'`, corresponding to the Gray labeling.

    Examples:
        The PAM modulation with order $M = 4$, base amplitude $A = 1$, and Gray labeling is depicted below.

        <figure markdown>
          ![4-PAM modulation with Gray labeling.](/figures/pam_4_gray.svg)
        </figure>

        >>> pam = komm.PAModulation(4)
        >>> pam.constellation
        array([-3., -1.,  1.,  3.])
        >>> pam.labeling
        array([[0, 0],
               [1, 0],
               [1, 1],
               [0, 1]])
        >>> pam.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1, 0])
        array([-3.,  1., -3., -1., -1.])
    """

    def __init__(
        self,
        order: int,
        base_amplitude: float = 1.0,
        labeling: Literal["natural", "reflected"] | npt.ArrayLike = "reflected",
    ) -> None:
        super().__init__(
            constellation=constellation_pam(order, base_amplitude),
            labeling=get_labeling(labeling, ("natural", "reflected"), order),
        )
        self.base_amplitude = base_amplitude
        self.labeling_parameter = labeling

    def __repr__(self) -> str:
        args = ", ".join([
            f"order={self.order}",
            f"base_amplitude={self.base_amplitude}",
            f"labeling='{self.labeling_parameter}'",
        ])
        return f"{self.__class__.__name__}({args})"

    @property
    @override
    def energy_per_symbol(self) -> float:
        return (self.base_amplitude**2) * (self.order**2 - 1) / 3

    @property
    @override
    def symbol_mean(self) -> float:
        return 0.0

    @property
    @override
    def minimum_distance(self) -> float:
        return 2.0 * self.base_amplitude

    @override
    def demodulate_hard(self, received: npt.ArrayLike) -> npt.NDArray[np.int_]:
        received = np.asarray(received)
        indices = np.clip(
            np.around((received + self.order - 1) / 2), 0, self.order - 1
        ).astype(int)
        hard_bits = np.reshape(self.labeling[indices], shape=-1)
        return hard_bits

    @override
    def demodulate_soft(
        self, received: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.float64]:
        print(self.order, self.labeling_parameter)
        if self.order == 2:
            y = np.asarray(received) / self.base_amplitude
            return -4 * snr * y  # [SA15, eq. (3.65)]
        elif self.order == 4 and self.labeling_parameter == "reflected":
            y = np.asarray(received) / self.base_amplitude
            return self._demodulate_pam4_soft_reflected(y, snr / 5.0)
        # Fall back to general implementation
        return super().demodulate_soft(received, snr)

    def _demodulate_pam4_soft_reflected(
        self, y: npt.NDArray[np.float64], gamma: float
    ) -> npt.NDArray[np.float64]:
        soft_bits = np.empty(2 * y.size, dtype=float)
        soft_bits[0::2] = (  # For bit_0: [SA15, eq. (5.15)]
            -8 * gamma + logcosh(6 * gamma * y) - logcosh(2 * gamma * y)
        )
        soft_bits[1::2] = (  # For bit_1: [SA15, eq. (5.6)]
            -8 * gamma * y + logcosh(2 * gamma * (y + 2)) - logcosh(2 * gamma * (y - 2))
        )
        return soft_bits



import numpy as np
import numpy.typing as npt

from ..._error_control_convolutional.TerminatedConvolutionalCode import (
    TerminatedConvolutionalCode,
)
from ..registry import RegistryBlockDecoder


def decode_bcjr(
    code: TerminatedConvolutionalCode,
    r: npt.ArrayLike,
    *,
    snr: float,
) -> npt.NDArray[np.float64]:
    if code.mode == "tail-biting":
        raise NotImplementedError("BCJR algorithm not implemented for 'tail-biting'")

    def metric_function(y: int, z: float) -> float:
        return 2.0 * snr * np.dot(code.cache_polar[y], z)

    n0 = code.convolutional_code.num_output_bits
    mu = code.convolutional_code.memory_order
    fsm = code.convolutional_code.finite_state_machine()

    if code.mode == "direct-truncation":
        initial_state_distribution = np.eye(1, fsm.num_states, 0)
        final_state_distribution = np.ones(fsm.num_states) / fsm.num_states
    else:  # code.mode == "zero-termination"
        initial_state_distribution = np.eye(1, fsm.num_states, 0)
        final_state_distribution = np.eye(1, fsm.num_states, 0)

    input_posteriors = fsm.forward_backward(
        observed_sequence=np.reshape(r, shape=(-1, n0)),
        metric_function=metric_function,
        initial_state_distribution=initial_state_distribution,
        final_state_distribution=final_state_distribution,
    )

    if code.mode == "zero-termination":
        input_posteriors = input_posteriors[:-mu]

    return np.log(input_posteriors[:, 0] / input_posteriors[:, 1])


RegistryBlockDecoder.register(
    "bcjr",
    {
        "description": "Bahl–Cocke–Jelinek–Raviv (BCJR)",
        "decoder": decode_bcjr,
        "type_in": "soft",
        "type_out": "soft",
        "target": "message",
    },
)



from typing import Any

import numpy as np
import numpy.typing as npt

from ..._algebra.BinaryPolynomial import BinaryPolynomial
from ..._algebra.FiniteBifield import F, FiniteBifield, FiniteBifieldElement, find_roots
from ..BCHCode import BCHCode
from ..registry import RegistryBlockDecoder


def bch_syndrome(
    code: BCHCode,
    r_poly: BinaryPolynomial,
) -> list[FiniteBifieldElement[FiniteBifield]]:
    # BCH syndrome computation. See [LC04, p. 205–209].
    alpha = code.field.primitive_element
    return [(r_poly % code.phi(i)).evaluate(alpha**i) for i in range(1, code.delta)]


def berlekamp_algorithm(
    field: F,
    delta: int,
    syndrome: list[Any],
) -> list[FiniteBifieldElement[F]]:
    # Berlekamp's iterative procedure for finding the error-location polynomial of a BCH code.
    # See [LC04, p. 209–212] and [RL09, p. 114–121].
    sigma = {-1: [field.one], 0: [field.one]}
    discrepancy = {-1: field.one, 0: syndrome[0]}
    degree = {-1: 0, 0: 0}

    for j in range(delta - 1):
        if discrepancy[j] == field.zero:
            degree[j + 1] = degree[j]
            sigma[j + 1] = sigma[j]
        else:
            rho, max_so_far = -1, -1
            for i in range(-1, j):
                if discrepancy[i] != field.zero and i - degree[i] > max_so_far:
                    rho, max_so_far = i, i - degree[i]
            degree[j + 1] = max(degree[j], degree[rho] + j - rho)
            sigma[j + 1] = [field.zero] * (degree[j + 1] + 1)
            first = np.array([field.zero] * (degree[j + 1] + 1), dtype=object)
            first[: degree[j] + 1] = sigma[j]
            second = np.array([field.zero] * (degree[j + 1] + 1), dtype=object)
            second[j - rho : degree[rho] + j - rho + 1] = sigma[rho]
            # [LC04, Eq. (6.25)]
            result = first + second * (discrepancy[j] / discrepancy[rho])
            sigma[j + 1] = result.tolist()
        if j < delta - 2:
            discrepancy[j + 1] = syndrome[j + 1]
            for i in range(degree[j + 1]):
                discrepancy[j + 1] += sigma[j + 1][i + 1] * syndrome[j - i]

    return sigma[delta - 1]


def decode_berlekamp(
    code: BCHCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    alpha = code.field.primitive_element
    r = np.asarray(r)
    r_poly = BinaryPolynomial.from_coefficients(r)
    syndrome = bch_syndrome(code, r_poly)
    if all(x == code.field.zero for x in syndrome):
        return r
    sigma_poly = berlekamp_algorithm(code.field, code.delta, syndrome)
    roots = find_roots(code.field, sigma_poly)
    e_loc = [e.inverse().logarithm(alpha) for e in roots]
    e_hat = np.bincount(e_loc, minlength=code.length)
    v_hat = (r + e_hat) % 2
    return v_hat


RegistryBlockDecoder.register(
    "berlekamp",
    {
        "description": "Berlekamp decoder",
        "decoder": decode_berlekamp,
        "type_in": "hard",
        "type_out": "hard",
        "target": "codeword",
    },
)



import functools
import inspect
from typing import Any, Callable, ParamSpec, Protocol, TypeVar

import numpy as np
import numpy.typing as npt
from attrs import asdict, make_class


class Attr(Protocol):
    name: str


def is_probability(inst: object, attr: Attr, value: float) -> None:
    if not 0 <= value <= 1:
        raise ValueError(f"'{attr.name}' must be between 0 and 1")


def is_pmf(inst: object, attr: Attr, value: npt.NDArray[np.float64]) -> None:
    if value.ndim != 1:
        raise ValueError(f"'{attr.name}' must be a 1D array")
    if not np.all(value >= 0.0):
        raise ValueError(f"'{attr.name}' must be non-negative")
    if not np.isclose(value.sum(), 1.0):
        raise ValueError(f"'{attr.name}' must sum to 1.0")


def is_transition_matrix(
    inst: object, attr: Attr, value: npt.NDArray[np.float64]
) -> None:
    if value.ndim != 2:
        raise ValueError(f"'{attr.name}' must be a 2D array")
    if not np.all(value >= 0.0):
        raise ValueError(f"'{attr.name}' must be non-negative")
    if not np.allclose(value.sum(axis=1), 1.0):
        raise ValueError(f"Rows of '{attr.name}' must sum to 1.0")


def is_log_base(inst: object, attr: Attr, value: float | str) -> None:
    if (isinstance(value, str) and value != "e") or (
        isinstance(value, float) and value <= 0.0
    ):
        raise ValueError(f"'{attr.name}' must be 'e' or a positive float")


P = ParamSpec("P")
R = TypeVar("R")


def validate_call(**fields: Any):
    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()
            Params = make_class("Params", fields)
            bound_args = {k: v for k, v in bound.arguments.items() if k in fields}
            inst = Params(**bound_args)  # Use attrs to validate and convert args
            bound.arguments.update(asdict(inst))
            return func(**bound.arguments)  # type: ignore

        return wrapper

    return decorator



from typing import Any

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._util.information_theory import LogBase, binary_entropy
from .._validation import is_pmf, is_probability, validate_call
from .AbstractDiscreteMemorylessChannel import AbstractDiscreteMemorylessChannel


@frozen
class BinarySymmetricChannel(AbstractDiscreteMemorylessChannel):
    r"""
    Binary symmetric channel (BSC). It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input and output alphabets $\mathcal{X} = \mathcal{Y} = \\{ 0, 1 \\}$. The channel is characterized by a parameter $p$, called the *crossover probability*. With probability $1 - p$, the output symbol is identical to the input symbol, and with probability $p$, the output symbol is flipped. Equivalently, the channel can be modeled as
    $$
        Y_n = X_n + Z_n,
    $$
    where $Z_n$ are iid Bernoulli random variables with $\Pr[Z_n = 1] = p$. For more details, see <cite>CT06, Sec. 7.1.4</cite>.

    Attributes:
        crossover_probability (Optional[float]): The channel crossover probability $p$. Must satisfy $0 \leq p \leq 1$. The default value is `0.0`, which corresponds to a noiseless channel.

    Parameters: Input:
        input_sequence (Array1D[int]): The input sequence.

    Parameters: Output:
        output_sequence (Array1D[int]): The output sequence.

    Examples:
        >>> np.random.seed(1)
        >>> bsc = komm.BinarySymmetricChannel(0.1)
        >>> bsc([0, 1, 1, 1, 0, 0, 0, 0, 0, 1])
        array([0, 1, 0, 1, 0, 1, 0, 0, 0, 1])
    """

    crossover_probability: float = field(default=0.0, validator=is_probability)

    @property
    def input_cardinality(self) -> int:
        return 2

    @property
    def output_cardinality(self) -> int:
        return 2

    @property
    def transition_matrix(self) -> npt.NDArray[np.float64]:
        r"""
        The transition probability matrix of the channel. It is given by
        $$
            p_{Y \mid X} = \begin{bmatrix} 1-p & p \\\\ p & 1-p \end{bmatrix}.
        $$

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.1)
            >>> bsc.transition_matrix
            array([[0.9, 0.1],
                   [0.1, 0.9]])
        """
        p = self.crossover_probability
        return np.array([[1 - p, p], [p, 1 - p]])

    @validate_call(input_pmf=field(converter=np.asarray, validator=is_pmf))
    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        Returns the mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$ of the channel. It is given by
        $$
            \mathrm{I}(X ; Y) = \Hb(p + \pi - 2 p \pi) - \Hb(p),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        **Parameters:**

        Same as the [corresponding method](/ref/DiscreteMemorylessChannel/#mutual_information) of the general class.

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.1)
            >>> bsc.mutual_information([0.45, 0.55])
            np.float64(0.5263828452309445)
        """
        input_pmf = np.array(input_pmf)
        p = self.crossover_probability
        pi = input_pmf[1]
        return (binary_entropy(p + pi - 2 * p * pi) - binary_entropy(p)) / np.log2(base)

    def capacity(self, base: LogBase = 2.0, **kwargs: Any) -> float:
        r"""
        Returns the channel capacity $C$. It is given by
        $$
            C = 1 - \Hb(p),
        $$
        in bits, where $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.1)
            >>> bsc.capacity()
            np.float64(0.5310044064107188)
        """
        p = self.crossover_probability
        return (1.0 - binary_entropy(p)) / np.log2(base)

    def __call__(self, input_sequence: npt.ArrayLike):
        p = self.crossover_probability
        input_sequence = np.array(input_sequence)
        error_pattern = (np.random.rand(np.size(input_sequence)) < p).astype(int)
        return (input_sequence + error_pattern) % 2



from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import numpy.typing as npt

from .._util.information_theory import LogBase


class AbstractDiscreteMemorylessChannel(ABC):
    @property
    @abstractmethod
    def transition_matrix(self) -> npt.NDArray[np.float64]:
        pass

    @property
    @abstractmethod
    def input_cardinality(self) -> int:
        pass

    @property
    @abstractmethod
    def output_cardinality(self) -> int:
        pass

    @abstractmethod
    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        pass

    @abstractmethod
    def capacity(self, base: LogBase = 2.0, **kwargs: Any) -> float:
        pass

    @abstractmethod
    def __call__(self, input_sequence: npt.ArrayLike) -> npt.NDArray[np.int_]:
        pass



import itertools as it
from functools import cached_property

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._algebra import BinaryPolynomial
from .._types import ArrayIntLike
from .BlockCode import BlockCode


@frozen(kw_only=True)
class CyclicCode(BlockCode):
    r"""
    General binary cyclic code. A cyclic code is a [linear block code](/ref/BlockCode) such that, if $c$ is a codeword, then every cyclic shift of $c$ is also a codeword. It is characterized by its *generator polynomial* $g(X)$, of degree $m$ (the redundancy of the code), and by its *check polynomial* $h(X)$, of degree $k$ (the dimension of the code). Those polynomials are related by $g(X) h(X) = X^n + 1$, where $n = k + m$ is the length of the code.

    Examples of generator polynomials can be found in the table below.

    | Code $(n, k, d)$  | Generator polynomial $g(X)$              | Integer representation           |
    | ----------------- | ---------------------------------------- | -------------------------------- |
    | Hamming $(7,4,3)$ | $X^3 + X + 1$                            | `0b1011 = 0o13 = 11`             |
    | Simplex $(7,3,4)$ | $X^4 + X^2 + X +   1$                    | `0b10111 = 0o27 = 23`            |
    | BCH $(15,5,7)$    | $X^{10} + X^8 + X^5 + X^4 + X^2 + X + 1$ | `0b10100110111 = 0o2467 = 1335`  |
    | Golay $(23,12,7)$ | $X^{11} + X^9 + X^7 + X^6 + X^5 + X + 1$ | `0b101011100011 = 0o5343 = 2787` |

    For more details, see <cite>LC04, Ch. 5</cite>.

    The constructor expects either the generator polynomial or the check polynomial.

    Parameters:
        length: The length $n$ of the code.

        generator_polynomial: The generator polynomial $g(X)$ of the code, of degree $m$ (the redundancy of the code), specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former.

        check_polynomial: The check polynomial $h(X)$ of the code, of degree $k$ (the dimension of the code), specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former.

        systematic: Whether the encoder is systematic. Default is `True`.

    Examples:
        >>> code = komm.CyclicCode(length=23, generator_polynomial=0b101011100011)  # Golay (23, 12)
        >>> (code.length, code.dimension, code.redundancy)
        (23, 12, 11)
        >>> code.minimum_distance()
        7

        >>> code = komm.CyclicCode(length=23, check_polynomial=0b1010010011111)  # Golay (23, 12)
        >>> (code.length, code.dimension, code.redundancy)
        (23, 12, 11)
        >>> code.minimum_distance()
        7
    """

    _length: int = field(default=None, repr=False, alias="length")
    _generator_polynomial: BinaryPolynomial | int = field(
        default=None, repr=False, alias="generator_polynomial"
    )
    _check_polynomial: BinaryPolynomial | int = field(
        default=None, repr=False, alias="check_polynomial"
    )
    _systematic: bool = field(default=True, repr=False, alias="systematic")

    def __attrs_post_init__(self) -> None:
        if (
            self._generator_polynomial is not None
            and self.modulus % self.generator_polynomial != 0b0
        ):
            raise ValueError("'generator_polynomial' must be a factor of X^n + 1")
        if (
            self._check_polynomial is not None
            and self.modulus % self.check_polynomial != 0b0
        ):
            raise ValueError("'check_polynomial' must be a factor of X^n + 1")

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(length={self.length}"
        if self._generator_polynomial is not None:
            s += f", generator_polynomial={self.generator_polynomial}"
        if self._check_polynomial is not None:
            s += f", check_polynomial={self.check_polynomial}"
        s += f", systematic={self.systematic})"
        return s

    @property
    def length(self) -> int:
        return self._length

    @cached_property
    def generator_polynomial(self) -> BinaryPolynomial:
        if self._generator_polynomial is None:
            return self.modulus // self.check_polynomial
        return BinaryPolynomial(self._generator_polynomial)

    @cached_property
    def check_polynomial(self) -> BinaryPolynomial:
        if self._check_polynomial is None:
            return self.modulus // self.generator_polynomial
        return BinaryPolynomial(self._check_polynomial)

    @property
    def systematic(self) -> bool:
        return self._systematic

    @cached_property
    def modulus(self) -> BinaryPolynomial:
        return BinaryPolynomial.from_exponents([0, self.length])

    @property
    def dimension(self) -> int:
        return self.check_polynomial.degree

    @property
    def redundancy(self) -> int:
        return self.generator_polynomial.degree

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        # See [LC04, Sec. 5.2].
        n, k, m = self.length, self.dimension, self.redundancy
        generator_matrix = np.empty((k, n), dtype=int)
        if not self.systematic:
            row = self.generator_polynomial.coefficients(width=n)
            for i in range(k):
                generator_matrix[i] = np.roll(row, i)
        else:
            generator_matrix[:, m:] = np.eye(k, dtype=int)
            for i in range(k):
                b_i_poly = (
                    BinaryPolynomial.from_exponents([m + i]) % self.generator_polynomial
                )
                b_i = b_i_poly.coefficients(width=m)
                generator_matrix[i, :m] = b_i
        return generator_matrix

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.int_]:
        # See [LC04, Sec. 5.2].
        if not self.systematic:
            n, m = self.length, self.redundancy
            check_matrix = np.empty((m, n), dtype=int)
            row = self.check_polynomial.coefficients(width=n)[::-1]
            for i in range(m):
                check_matrix[m - i - 1] = np.roll(row, -i)
            return check_matrix
        else:
            raise NotImplementedError

    def enc_mapping(self, u: ArrayIntLike) -> npt.NDArray[np.int_]:
        u_poly = BinaryPolynomial.from_coefficients(u)
        if not self.systematic:
            v_poly = u_poly * self.generator_polynomial
        else:
            u_poly_shifted = u_poly << self.redundancy
            b_poly = u_poly_shifted % self.generator_polynomial
            v_poly = u_poly_shifted + b_poly
        return v_poly.coefficients(width=self.length)

    def inv_enc_mapping(self, v: ArrayIntLike) -> npt.NDArray[np.int_]:
        if not self.systematic:
            v_poly = BinaryPolynomial.from_coefficients(v)
            u_poly = v_poly // self.generator_polynomial
            u = u_poly.coefficients(width=self.dimension)
        else:
            u = np.take(v, range(self.redundancy, self.length))
        return u

    def chk_mapping(self, r: ArrayIntLike) -> npt.NDArray[np.int_]:
        r_poly = BinaryPolynomial.from_coefficients(r)
        s_poly = r_poly % self.generator_polynomial
        s = s_poly.coefficients(width=self.redundancy)
        return s

    @property
    def default_decoder(self) -> str:
        return "meggitt"

    @classmethod
    def supported_decoders(cls) -> list[str]:
        return cls.__base__.supported_decoders() + ["meggitt"]  # type: ignore

    @cached_property
    def meggitt_table(self) -> dict[BinaryPolynomial, BinaryPolynomial]:
        r"""
        The Meggitt table for the cyclic code. It is a dictionary where the keys are syndromes and the values are error patterns. See <cite>XiD03, Sec. 3.4</cite>.
        """
        meggitt_table: dict[BinaryPolynomial, BinaryPolynomial] = {}
        for w in range(self.packing_radius() + 1):
            for idx in it.combinations(range(self.length - 1), w):
                e_poly = BinaryPolynomial.from_exponents(list(idx) + [self.length - 1])
                s_poly = e_poly % self.generator_polynomial
                meggitt_table[s_poly] = e_poly
        return meggitt_table



from itertools import combinations, product

import numpy as np
import numpy.typing as npt
from attrs import define, field

from .._types import ArrayIntLike


@define
class Modulation:
    r"""
    General modulation scheme. A *modulation scheme* of *order* $M = 2^m$ is defined by a *constellation* $\mathbf{X}$, which is a real or complex vector of length $M$, and a *binary labeling* $\mathbf{Q}$, which is an $M \times m$ binary matrix whose rows are all distinct. The $i$-th element of $\mathbf{X}$, for $i \in [0:M)$, is denoted by $x_i$ and is called the $i$-th *constellation symbol*. The $i$-th row of $\mathbf{Q}$, for $i \in [0:M)$, is called the *binary representation* of the $i$-th constellation symbol. For more details, see <cite>SA15, Sec. 2.5</cite>.

    Parameters:
        constellation (Array1D[float] | Array1D[complex]): The constellation $\mathbf{X}$ of the modulation. Must be a 1D-array containing $M$ real or complex numbers.

        labeling (Array2D[int]): The binary labeling $\mathbf{Q}$ of the modulation. Must be a 2D-array of shape $(M, m)$ where each row is a distinct binary $m$-tuple.

    Examples:
        The real modulation scheme depicted in the figure below has $M = 4$ and $m = 2$.

        <figure markdown>
          ![Example for real modulation with M = 4](/figures/modulation_real_4.svg)
        </figure>

        The constellation is given by
        $$
          \mathbf{X} = \begin{bmatrix}
            -0.5 \\\\
             0.0 \\\\
             0.5 \\\\
             2.0
          \end{bmatrix},
        $$
        and the binary labeling is given by
        $$
          \mathbf{Q} = \begin{bmatrix}
            1 & 0 \\\\
            1 & 1 \\\\
            0 & 1 \\\\
            0 & 0
          \end{bmatrix}.
        $$

        >>> modulation = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
        >>> modulation.constellation
        array([-0.5,  0. ,  0.5,  2. ])
        >>> modulation.labeling
        array([[1, 0],
               [1, 1],
               [0, 1],
               [0, 0]])

        The complex modulation scheme depicted in the figure below has $M = 4$ and $m = 2$.

        <figure markdown>
          ![Example for complex modulation with M = 4](/figures/modulation_complex_4.svg)
        </figure>

        The constellation is given by
        $$
          \mathbf{X} = \begin{bmatrix}
             0 \\\\
            -1 \\\\
             1 \\\\
            \mathrm{j}
          \end{bmatrix},
        $$
        and the binary labeling is given by
        $$
          \mathbf{Q} = \begin{bmatrix}
            0 & 0 \\\\
            0 & 1 \\\\
            1 & 0 \\\\
            1 & 1
          \end{bmatrix}.
        $$

        >>> modulation = komm.Modulation(constellation=[0, -1, 1, 1j], labeling=[[0, 0], [0, 1], [1, 0], [1, 1]])
        >>> modulation.constellation
        array([ 0.+0.j, -1.+0.j,  1.+0.j,  0.+1.j])
        >>> modulation.labeling
        array([[0, 0],
               [0, 1],
               [1, 0],
               [1, 1]])
    """

    constellation: npt.NDArray[np.float64 | np.complex128] = field(
        converter=np.asarray,
        repr=lambda x: x.tolist(),
    )
    labeling: npt.NDArray[np.int_] = field(
        converter=np.asarray,
        repr=lambda x: x.tolist(),
    )

    def __attrs_post_init__(self) -> None:
        order, m = self.order, self.bits_per_symbol
        if order & (order - 1):
            raise ValueError("length of 'constellation' must be a power of two")
        if self.labeling.shape != (order, m):
            raise ValueError(f"shape of 'labeling' must be ({order}, {m})")
        if np.any(self.labeling < 0) or np.any(self.labeling > 1):
            raise ValueError("elements of 'labeling' must be either 0 or 1")
        if len(set(tuple(row) for row in self.labeling)) != order:
            raise ValueError("rows of 'labeling' must be distinct")

    @property
    def order(self) -> int:
        r"""
        The order $M$ of the modulation.

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.order
            4
        """
        return self.constellation.size

    @property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        The inverse labeling of the modulation. It is a dictionary that maps each binary tuple to the corresponding constellation index.

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.inverse_labeling
            {(1, 0): 0, (1, 1): 1, (0, 1): 2, (0, 0): 3}
        """
        return {tuple(int(x) for x in row): i for i, row in enumerate(self.labeling)}

    @property
    def bits_per_symbol(self) -> int:
        r"""
        The number $m$ of bits per symbol of the modulation. It is given by $m = \log_2 M$, where $M$ is the order of the modulation.

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.bits_per_symbol
            2
        """
        return (self.order - 1).bit_length()

    @property
    def energy_per_symbol(self) -> float:
        r"""
        The average symbol energy $E_\mathrm{s}$ of the constellation. It assumes equiprobable symbols. It is given by
        $$
          E_\mathrm{s} = \frac{1}{M} \sum_{i \in [0:M)} |x_i|^2,
        $$
        where $|x_i|^2$ is the energy of constellation symbol $x_i$, and $M$ is the order of the modulation.

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.energy_per_symbol
            np.float64(1.125)

            >>> mod = komm.Modulation(constellation=[0, -1, 1, 1j], labeling=[[0, 0], [0, 1], [1, 0], [1, 1]])
            >>> mod.energy_per_symbol
            np.float64(0.75)
        """
        return ((self.constellation * self.constellation.conj()).real).mean()

    @property
    def energy_per_bit(self) -> float:
        r"""
        The average bit energy $E_\mathrm{b}$ of the constellation. It assumes equiprobable symbols. It is given by $E_\mathrm{b} = E_\mathrm{s} / m$, where $E_\mathrm{s}$ is the average symbol energy, and $m$ is the number of bits per symbol of the modulation.

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.energy_per_bit
            np.float64(0.5625)

            >>> mod = komm.Modulation(constellation=[0, -1, 1, 1j], labeling=[[0, 0], [0, 1], [1, 0], [1, 1]])
            >>> mod.energy_per_bit
            np.float64(0.375)
        """
        return self.energy_per_symbol / self.bits_per_symbol

    @property
    def symbol_mean(self) -> float | complex:
        r"""
        The mean $\mu_\mathrm{s}$ of the constellation. It assumes equiprobable symbols. It is given by
        $$
          \mu_\mathrm{s} = \frac{1}{M} \sum_{i \in [0:M)} x_i.
        $$

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.symbol_mean
            np.float64(0.5)

            >>> mod = komm.Modulation(constellation=[0, -1, 1, 1j], labeling=[[0, 0], [0, 1], [1, 0], [1, 1]])
            >>> mod.symbol_mean
            np.complex128(0.25j)
        """
        return self.constellation.mean()

    @property
    def minimum_distance(self) -> float:
        r"""
        The minimum Euclidean distance $d_\mathrm{min}$ of the constellation. It is given by
        $$
            d_\mathrm{min} = \min_{i, j \in [0:M), ~ i \neq j} |x_i - x_j|.
        $$

        Examples:
            >>> mod = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> mod.minimum_distance
            np.float64(0.5)

            >>> mod = komm.Modulation(constellation=[0, -1, 1, 1j], labeling=[[0, 0], [0, 1], [1, 0], [1, 1]])
            >>> mod.minimum_distance
            np.float64(1.0)
        """
        return np.min(
            [np.abs(s1 - s2) for s1, s2 in combinations(self.constellation, 2)]
        )

    def modulate(self, bits: ArrayIntLike) -> npt.NDArray[np.float64 | np.complex128]:
        r"""
        Modulates a sequence of bits to its corresponding constellation symbols.

        Parameters:
            bits (Array1D[int]): The bits to be modulated. It should be a 1D-array of integers in the set $\{ 0, 1 \}$. Its length must be a multiple of $m$.

        Returns:
            symbols (Array1D[complex] | Array1D[float]): The constellation symbols corresponding to `bits`. It is a 1D-array of real or complex numbers. Its length is equal to the length of `bits` divided by $m$.

        Examples:
            >>> modulation = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> modulation.modulate([0, 0, 1, 1, 0, 0, 1, 0])
            array([ 2. ,  0. ,  2. , -0.5])
        """
        bits = np.asarray(bits, dtype=int)
        m = self.bits_per_symbol
        n_symbols = bits.size // m
        if bits.size != n_symbols * m:
            raise ValueError("length of 'bits' must be a multiple of bits per symbol")
        symbols = np.empty(n_symbols, dtype=self.constellation.dtype)
        for i, bit_sequence in enumerate(np.reshape(bits, shape=(n_symbols, m))):
            symbols[i] = self.constellation[self.inverse_labeling[tuple(bit_sequence)]]
        return symbols

    def demodulate_hard(self, received: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Demodulates a sequence of received points to a sequence of bits using hard-decision decoding.

        Parameters:
            received (Array1D[T]): The received points to be demodulated. It should be a 1D-array of real or complex numbers. It may be of any length.

        Returns:
            hard_bits (Array1D[int]): The bits corresponding to `received`. It is a 1D-array of bits (integers in the set $\{ 0, 1 \}$). Its length is equal to the length of `received` multiplied by $m$.

        Examples:
            >>> modulation = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> modulation.demodulate_hard([2.17, -0.06, 1.94, -0.61])
            array([0, 0, 1, 1, 0, 0, 1, 0])
        """
        # General minimum Euclidean distance hard demodulator.
        received = np.asarray(received)
        hard_bits = np.empty((received.size, self.bits_per_symbol), dtype=int)
        for i, y in enumerate(received):
            hard_bits[i, :] = self.labeling[
                np.argmin(np.abs(self.constellation - y)), :
            ]
        hard_bits = np.reshape(hard_bits, shape=-1)
        return hard_bits

    def demodulate_soft(
        self, received: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.float64]:
        r"""
        Demodulates a sequence of received points to a sequence of bits using soft-decision decoding.

        Parameters:
            received (Array1D[T]): The received points to be demodulated. It should be a 1D-array of real or complex numbers. It may be of any length.

            snr (float): The signal-to-noise ratio (SNR) of the channel. It should be a positive real number.

        Returns:
            soft_bits (Array1D[float]): The soft bits corresponding to `received`. It is a 1D-array of L-values (real numbers, where positive values correspond to bit $0$ and negative values correspond to bit $1$). Its length is equal to the length of `received` multiplied by $m$.

        Examples:
            >>> modulation = komm.Modulation(constellation=[-0.5, 0.0, 0.5, 2.0], labeling=[[1, 0], [1, 1], [0, 1], [0, 0]])
            >>> modulation.demodulate_soft([2.17, -0.06, 1.94, -0.61], snr=100.0)
            array([ 416.        ,  245.33333333,  -27.5555556 ,  -16.88888889,
                    334.22222222,  184.        , -108.44444444,   32.        ])
        """
        # Computes the L-values (LLR) of each bit. Assumes uniformly distributed bits.
        # See [SA15, eq. (3.50)].
        m = self.bits_per_symbol
        n0 = self.energy_per_symbol / snr

        def pdf_received_given_bit(bit_index: int, bit_value: int) -> float:
            bits = np.empty(m, dtype=int)
            bits[bit_index] = bit_value
            rest_index = np.setdiff1d(np.arange(m), [bit_index])
            f = 0.0
            for b_rest in product([0, 1], repeat=m - 1):
                bits[rest_index] = b_rest
                point = self.constellation[self.inverse_labeling[tuple(bits)]]
                f += np.exp(-np.abs(received - point) ** 2 / n0)
            return f

        received = np.asarray(received)
        soft_bits = np.empty(received.size * m, dtype=float)
        for bit_index in range(m):
            p0 = pdf_received_given_bit(bit_index, 0)
            p1 = pdf_received_given_bit(bit_index, 1)
            with np.errstate(divide="ignore"):
                soft_bits[bit_index::m] = np.log(p0) - np.log(p1)
        return soft_bits



from .bit_operations import binlist2int, int2binlist, pack, unpack
from .correlation import acorr, cyclic_acorr
from .information_theory import binary_entropy, entropy, relative_entropy
from .special_functions import qfunc, qfuncinv

__all__ = [
    "binlist2int",
    "int2binlist",
    "pack",
    "unpack",
    "acorr",
    "cyclic_acorr",
    "entropy",
    "binary_entropy",
    "relative_entropy",
    "qfunc",
    "qfuncinv",
]



import numpy as np
from attrs import frozen
from numpy import typing as npt

from .BlockCode import BlockCode


@frozen
class BlockEncoder:
    r"""
    Encoder for [linear block codes](/ref/BlockCode).

    Attributes:
        code: The [block code](/ref/BlockCode) to be considered.

    Parameters: Input:
        in0 (Array1D[int]): The bit sequence to be encoded. Its length must be a multiple of the code's dimension $k$.

    Parameters: Output:
        out0 (Array1D[int]): The encoded bit sequence. Its length is a multiple of the code's block length $n$.

    Examples:
        >>> code = komm.HammingCode(3)
        >>> encoder = komm.BlockEncoder(code)
        >>> encoder([1, 1, 0, 0, 1, 0, 1, 1])
        array([1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0])
    """

    code: BlockCode

    def __call__(self, in0: npt.ArrayLike) -> npt.NDArray[np.int_]:
        u = np.reshape(in0, (-1, self.code.dimension))
        v = np.apply_along_axis(self.code.enc_mapping, 1, u)
        out0 = np.reshape(v, (-1,))
        return out0



from functools import partial
from typing import Any, Literal

import numpy as np
from attrs import field, frozen
from numpy import typing as npt

from .BlockCode import BlockCode
from .registry import RegistryBlockDecoder


@frozen
class BlockDecoder:
    r"""
    Decoder for [linear block codes](/ref/BlockCode).

    Attributes:
        code: The [block code](/ref/BlockCode) to be considered.
        method: The identifier of the method to be used for decoding. The default corresponds to `code.default_decoder`.
        decoder_kwargs: Additional keyword arguments to be passed to the decoder.

    Notes:
        - To see the default decoding method for a given code, use `code.default_decoder`; to see the available decoding methods for a given code, use `code.supported_decoders()`.

    ??? info "Available decoding methods"

        **`bcjr`**: Bahl–Cocke–Jelinek–Raviv (BCJR)

        - Input type: soft
        - Output type: soft
        - Target: message
        - Supported by: [`TerminatedConvolutionalCode`](/ref/TerminatedConvolutionalCode).

        **`berlekamp`**: Berlekamp decoder

        - Input type: hard
        - Output type: hard
        - Target: codeword
        - Supported by: [`BCHCode`](/ref/BCHCode).

        **`exhaustive-search-hard`**: Exhaustive search (hard-decision). Minimum Hamming distance decoder

        - Input type: hard
        - Output type: hard
        - Target: codeword
        - Supported by: All codes.

        **`exhaustive-search-soft`**: Exhaustive search (soft-decision). Minimum Euclidean distance decoder

        - Input type: soft
        - Output type: hard
        - Target: codeword
        - Supported by: All codes.

        **`majority-logic-repetition-code`**: Majority-logic decoder. A hard-decision decoder for Repetition codes only.

        - Input type: hard
        - Output type: hard
        - Target: message
        - Supported by: [`RepetitionCode`](/ref/RepetitionCode).

        **`meggitt`**: Meggitt decoder

        - Input type: hard
        - Output type: hard
        - Target: codeword
        - Supported by: [`BCHCode`](/ref/BCHCode), [`CyclicCode`](/ref/CyclicCode).

        **`reed`**: Reed decoding algorithm for Reed–Muller codes. It's a majority-logic decoding algorithm.

        - Input type: hard
        - Output type: hard
        - Target: message
        - Supported by: [`ReedMullerCode`](/ref/ReedMullerCode).

        **`syndrome-table`**: Syndrome table decoder

        - Input type: hard
        - Output type: hard
        - Target: codeword
        - Supported by: All codes.

        **`viterbi-hard`**: Viterbi (hard-decision)

        - Input type: hard
        - Output type: hard
        - Target: message
        - Supported by: [`TerminatedConvolutionalCode`](/ref/TerminatedConvolutionalCode).

        **`viterbi-soft`**: Viterbi (soft-decision)

        - Input type: soft
        - Output type: hard
        - Target: message
        - Supported by: [`TerminatedConvolutionalCode`](/ref/TerminatedConvolutionalCode).

        **`wagner`**: Wagner decoder. A soft-decision decoder for SingleParityCheck codes only.

        - Input type: soft
        - Output type: hard
        - Target: codeword
        - Supported by: [`SingleParityCheckCode`](/ref/SingleParityCheckCode).

        **`weighted-reed`**: Weighted Reed decoding algorithm for Reed–Muller codes.

        - Input type: soft
        - Output type: hard
        - Target: message
        - Supported by: [`ReedMullerCode`](/ref/ReedMullerCode).

    Parameters: Input:
        in0 (Array1D[int] | Array1D[float]): The (hard or soft) bit sequence to be decoded. Its length must be a multiple of the code's block length $n$.

    Parameters: Output:
        out0 (Array1D[int]): The decoded bit sequence. Its length is a multiple of the code's dimension $k$.

    Examples:
        >>> code = komm.HammingCode(3)
        >>> code.default_decoder
        'syndrome-table'
        >>> code.supported_decoders()
        ['exhaustive-search-hard', 'exhaustive-search-soft', 'syndrome-table']

        >>> decoder = komm.BlockDecoder(code)
        >>> decoder([1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1, 0, 0, 0])
        array([1, 1, 0, 0, 1, 0, 1, 1])

        >>> decoder = komm.BlockDecoder(code, method='exhaustive-search-soft')
        >>> decoder([-0.98, -0.85, 1.07, -0.78, 1.11, -0.95, -1.16, -0.87, 1.11, -0.83, -0.95, 0.94, 1.07, 0.91])
        array([1, 1, 0, 0, 1, 0, 1, 1])
    """

    code: BlockCode
    method: (
        Literal[
            "bcjr",
            "berlekamp",
            "exhaustive-search-hard",
            "exhaustive-search-soft",
            "majority-logic-repetition-code",
            "meggitt",
            "reed",
            "syndrome-table",
            "viterbi-hard",
            "viterbi-soft",
            "wagner",
            "weighted-reed",
        ]
        | None
    ) = None
    decoder_kwargs: dict[str, Any] = field(factory=dict)

    def __attrs_post_init__(self) -> None:
        method = self.method or self.code.default_decoder
        supported_methods_info = (
            f"supported methods for {self.code.__class__.__name__}: "
            f"{set(self.code.supported_decoders())}"
        )
        if not RegistryBlockDecoder.is_registered(method):
            error_message = f"method '{method}' is unknown"
            raise ValueError(f"{error_message}\n{supported_methods_info}")
        if method not in self.code.supported_decoders():
            error_message = f"method '{method}' is not supported for this code"
            raise ValueError(f"{error_message}\n{supported_methods_info}")

    def __call__(self, in0: npt.ArrayLike) -> npt.NDArray[np.int_ | np.float64]:
        method = self.method or self.code.default_decoder
        decoder_data = RegistryBlockDecoder.get(method)

        if decoder_data["type_in"] == "hard" and not np.issubdtype(
            np.asarray(in0).dtype, np.integer
        ):
            raise TypeError(f"input type must be 'int' for method '{method}'")
        if decoder_data["type_in"] == "soft" and not np.issubdtype(
            np.asarray(in0).dtype, np.floating
        ):
            raise TypeError(f"input type must be 'float' for method '{method}'")

        decoded = np.apply_along_axis(
            func1d=partial(decoder_data["decoder"], self.code, **self.decoder_kwargs),
            axis=1,
            arr=np.reshape(in0, (-1, self.code.length)),
        )

        if decoder_data["target"] == "codeword":
            u_hat = np.apply_along_axis(self.code.inv_enc_mapping, 1, decoded)
        else:
            u_hat = decoded

        out0 = np.reshape(u_hat, (-1,))
        return out0



from .ConvolutionalCode import ConvolutionalCode
from .ConvolutionalStreamDecoder import ConvolutionalStreamDecoder
from .ConvolutionalStreamEncoder import ConvolutionalStreamEncoder
from .PuncturedConvolutionalCode import PuncturedConvolutionalCode
from .TerminatedConvolutionalCode import TerminatedConvolutionalCode

__all__ = [
    "ConvolutionalCode",
    "PuncturedConvolutionalCode",
    "TerminatedConvolutionalCode",
    "ConvolutionalStreamDecoder",
    "ConvolutionalStreamEncoder",
]



from typing import Optional, cast

import numpy as np
import numpy.typing as npt
from attrs import field

from .._validation import is_pmf, validate_call
from .util import tunstall_algorithm
from .VariableToFixedCode import VariableToFixedCode


@validate_call(
    pmf=field(converter=np.asarray, validator=is_pmf),
)
def TunstallCode(
    pmf: npt.ArrayLike,
    target_block_size: Optional[int] = None,
) -> VariableToFixedCode:
    r"""
    Binary Tunstall code. It is an optimal (minimal expected rate) [variable-to-fixed length code](/ref/VariableToFixedCode) for a given probability mass function. For more details, see <cite>Say06, Sec. 3.7</cite>.

    Parameters:
        pmf: The probability mass function of the source.
        target_block_size: The target block size $n$. Must satisfy $2^n \geq S$, where $S$ is the cardinality of the source alphabet, given by `len(pmf)`. The default value is $n = \lceil \log_2 S \rceil$.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.TunstallCode(pmf)
        >>> code.dec_mapping  # doctest: +NORMALIZE_WHITESPACE
        {(0, 0): (0,),
         (0, 1): (1,),
         (1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +NUMBER
        np.float64(2.0)

        >>> code = komm.TunstallCode(pmf, 3)
        >>> code.dec_mapping  # doctest: +NORMALIZE_WHITESPACE
        {(0, 0, 0): (0, 0, 0),
         (0, 0, 1): (0, 0, 1),
         (0, 1, 0): (0, 0, 2),
         (0, 1, 1): (0, 1),
         (1, 0, 0): (0, 2),
         (1, 0, 1): (1,),
         (1, 1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +NUMBER
        np.float64(1.3698630137)
    """
    pmf = cast(npt.NDArray[np.float64], pmf)
    if target_block_size is None:
        target_block_size = int(np.ceil(np.log2(pmf.size)))
    elif 2**target_block_size < pmf.size:
        raise ValueError("'target_block_size' is too low")
    return VariableToFixedCode.from_sourcewords(
        target_cardinality=2,
        sourcewords=tunstall_algorithm(pmf, target_block_size),
    )



import numpy as np
import numpy.typing as npt

from ..._algebra import BinaryPolynomial
from ..CyclicCode import CyclicCode
from ..registry import RegistryBlockDecoder


def decode_meggitt(
    code: CyclicCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    # See [XiD03, Sec. 3.4] for more details.
    r = np.asarray(r)
    meggitt_table = code.meggitt_table
    r_poly = BinaryPolynomial.from_coefficients(r)
    s_poly = r_poly % code.generator_polynomial
    if s_poly == 0:
        return np.asarray(r)
    e_poly_hat = BinaryPolynomial(0)
    for j in range(code.length):
        if s_poly in meggitt_table:
            e_poly_hat = meggitt_table[s_poly] // BinaryPolynomial(1 << j)
            break
        s_poly = (s_poly << 1) % code.generator_polynomial
    return (r_poly + e_poly_hat).coefficients(code.length)


RegistryBlockDecoder.register(
    "meggitt",
    {
        "description": "Meggitt decoder",
        "decoder": decode_meggitt,
        "type_in": "hard",
        "type_out": "hard",
        "target": "codeword",
    },
)



import heapq
import itertools as it
import math
from typing import Literal

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

Word = tuple[int, ...]


def is_prefix_free(words: list[Word]) -> bool:
    for c1, c2 in it.permutations(words, 2):
        if c1[: len(c2)] == c2:
            return False
    return True


def is_uniquely_decodable(words: list[Word]) -> bool:
    # Sardinas–Patterson algorithm. See [Say06, Sec. 2.4.1].
    augmented_words = set(words)
    while True:
        dangling_suffixes: set[Word] = set()
        for c1, c2 in it.permutations(augmented_words, 2):
            if c1[: len(c2)] == c2:
                dangling_suffixes.add(c1[len(c2) :])
        if dangling_suffixes & set(words):
            return False
        if dangling_suffixes <= augmented_words:
            return True
        augmented_words |= dangling_suffixes


def parse_prefix_free(
    input_sequence: npt.NDArray[np.int_], dictionary: dict[Word, Word]
) -> npt.NDArray[np.int_]:
    output_sequence: list[int] = []
    i = 0
    while i < len(input_sequence):
        j = 1
        while i + j <= len(input_sequence):
            try:
                key = tuple(input_sequence[i : i + j])
                output_sequence.extend(dictionary[key])
                break
            except KeyError:
                j += 1
        i += j
    return np.asarray(output_sequence)


def huffman_algorithm(
    pmf: npt.NDArray[np.float64], source_block_size: int, policy: Literal["high", "low"]
) -> list[Word]:
    class Node:
        def __init__(self, index: int, probability: float):
            self.index: int = index
            self.probability: float = probability
            self.parent: int | None = None
            self.bit: int = -1

        def __lt__(self, other: Self) -> bool:
            i0, p0 = self.index, self.probability
            i1, p1 = other.index, other.probability
            if policy == "high":
                return (p0, i0) < (p1, i1)
            elif policy == "low":
                return (p0, -i0) < (p1, -i1)

    tree = [
        Node(i, math.prod(probs))
        for (i, probs) in enumerate(it.product(pmf, repeat=source_block_size))
    ]
    queue = [node for node in tree]
    heapq.heapify(queue)
    while len(queue) > 1:
        node1 = heapq.heappop(queue)
        node0 = heapq.heappop(queue)
        node1.bit = 1
        node0.bit = 0
        node = Node(index=len(tree), probability=node0.probability + node1.probability)
        node0.parent = node1.parent = node.index
        heapq.heappush(queue, node)
        tree.append(node)

    codewords: list[Word] = []
    for symbol in range(pmf.size**source_block_size):
        node = tree[symbol]
        bits: list[int] = []
        while node.parent is not None:
            bits.insert(0, node.bit)
            node = tree[node.parent]
        codewords.append(tuple(bits))

    return codewords


def tunstall_algorithm(
    pmf: npt.NDArray[np.float64], code_block_size: int
) -> list[Word]:
    class Node:
        def __init__(self, symbols: Word, probability: float):
            self.symbols = symbols
            self.probability = probability

        def __lt__(self, other: Self) -> bool:
            return -self.probability < -other.probability

    queue = [Node((symbol,), probability) for (symbol, probability) in enumerate(pmf)]
    heapq.heapify(queue)

    while len(queue) + pmf.size - 1 < 2**code_block_size:
        node = heapq.heappop(queue)
        for symbol, probability in enumerate(pmf):
            new_node = Node(node.symbols + (symbol,), node.probability * probability)
            heapq.heappush(queue, new_node)
    sourcewords = sorted(node.symbols for node in queue)

    return sourcewords



from typing import Any

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._util.information_theory import LogBase, arimoto_blahut, mutual_information
from .._validation import is_transition_matrix
from .AbstractDiscreteMemorylessChannel import AbstractDiscreteMemorylessChannel


@frozen
class DiscreteMemorylessChannel(AbstractDiscreteMemorylessChannel):
    r"""
    General discrete memoryless channel (DMC). It is defined by an *input alphabet* $\mathcal{X}$, an *output alphabet* $\mathcal{Y}$, and a *transition probability matrix* $p_{Y \mid X}$. Here, for simplicity, the input and output alphabets are always taken as $\mathcal{X} = \\{ 0, 1, \ldots, |\mathcal{X}| - 1 \\}$ and $\mathcal{Y} = \\{ 0, 1, \ldots, |\mathcal{Y}| - 1 \\}$, respectively. The transition probability matrix $p_{Y \mid X}$, of size $|\mathcal{X}|$-by-$|\mathcal{Y}|$, gives the conditional probability of receiving $Y = y$ given that $X = x$ is transmitted. For more details, see <cite>CT06, Ch. 7</cite>.

    Attributes:
        transition_matrix (Array2D[float]): The channel transition probability matrix $p_{Y \mid X}$. The element in row $x \in \mathcal{X}$ and column $y \in \mathcal{Y}$ must be equal to $p_{Y \mid X}(y \mid x)$.

    Parameters: Input:
        input_sequence (Array1D[int]): The input sequence.

    Parameters: Output:
        output_sequence (Array1D[int]): The output sequence.

    Examples:
        >>> np.random.seed(1)
        >>> dmc = komm.DiscreteMemorylessChannel([[0.9, 0.05, 0.05], [0.0, 0.5, 0.5]])
        >>> dmc([0, 1, 0, 1, 1, 1, 0, 0, 0, 1])
        array([0, 2, 0, 1, 1, 1, 0, 0, 0, 2])

    """

    transition_matrix: npt.NDArray[np.float64] = field(
        converter=np.asarray, validator=is_transition_matrix
    )

    @property
    def input_cardinality(self) -> int:
        r"""
        The channel input cardinality $|\mathcal{X}|$.
        """
        return self.transition_matrix.shape[0]

    @property
    def output_cardinality(self) -> int:
        r"""
        The channel output cardinality $|\mathcal{Y}|$.
        """
        return self.transition_matrix.shape[1]

    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        Returns the mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$ of the channel. It is given by
        $$
            \mathrm{I}(X ; Y) = \mathrm{H}(X) - \mathrm{H}(X \mid Y),
        $$
        where $\mathrm{H}(X)$ is the the entropy of $X$ and $\mathrm{H}(X \mid Y)$ is the conditional entropy of $X$ given $Y$. By default, the base of the logarithm is $2$, in which case the mutual information is measured in bits. See <cite>CT06, Ch. 2</cite>.

        Parameters:
            input_pmf (Array1D[float]): The probability mass function $p_X$ of the channel input $X$. It must be a valid pmf, that is, all of its values must be non-negative and sum up to $1$.

            base (Optional[float | str]): The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

        Returns:
            mutual_information (float): The mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$.

        Examples:
            >>> dmc = komm.DiscreteMemorylessChannel([[0.6, 0.3, 0.1], [0.7, 0.1, 0.2], [0.5, 0.05, 0.45]])
            >>> dmc.mutual_information([1/3, 1/3, 1/3]).round(6)
            np.float64(0.123811)
            >>> dmc.mutual_information([1/3, 1/3, 1/3], base=3).round(6)
            np.float64(0.078116)
            >>> dmc.mutual_information([1/3, 1/3, 1/3], base='e').round(6)
            np.float64(0.085819)
        """
        return mutual_information(input_pmf, self.transition_matrix, base)

    def capacity(self, base: LogBase = 2.0, **kwargs: Any) -> float:
        r"""
        Returns the channel capacity $C$. It is given by $C = \max_{p_X} \mathrm{I}(X;Y)$. This method computes the channel capacity via the Arimoto–Blahut algorithm. See <cite>CT06, Sec. 10.8</cite>.

        Parameters:
            base (Optional[float | str]): The base of the logarithm to be used. It must be a positive float or the string `'e'`. The default value is `2.0`.

        Parameters: Additional keyword arguments for the Arimoto–Blahut algorithm:
            max_iter (Optional[int]): The maximum number of iterations. The default value is `1000`.
            tol (Optional[float]): The error tolerance. The default value is `1e-12`.

        Returns:
            capacity (float): The channel capacity $C$.

        Examples:
            >>> dmc = komm.DiscreteMemorylessChannel([[0.6, 0.3, 0.1], [0.7, 0.1, 0.2], [0.5, 0.05, 0.45]])
            >>> dmc.capacity().round(6)
            np.float64(0.161632)
            >>> dmc.capacity(base=3).round(6)
            np.float64(0.101978)
            >>> dmc.capacity(base='e').round(6)
            np.float64(0.112035)
        """
        if not kwargs:
            kwargs = {"max_iter": 1000, "tol": 1e-12}
        initial_guess = np.ones(self.input_cardinality) / self.input_cardinality
        optimal_input_pmf = arimoto_blahut(
            self.transition_matrix, initial_guess, **kwargs
        )
        return mutual_information(optimal_input_pmf, self.transition_matrix, base=base)

    def __call__(self, input_sequence: npt.ArrayLike) -> npt.NDArray[np.int_]:
        input_sequence = np.asarray(input_sequence)
        output_sequence = [
            np.random.choice(self.output_cardinality, p=self.transition_matrix[x])
            for x in input_sequence
        ]
        return np.array(output_sequence)



from statistics import NormalDist

import numpy as np
import numpy.typing as npt

norm = NormalDist()


def _qfunc(x: float) -> float:
    return norm.cdf(-x)


def _qfuncinv(y: float) -> float:
    return -norm.inv_cdf(y) + 0.0  # + 0.0 to avoid -0.0


def qfunc(x: npt.ArrayLike) -> npt.NDArray[np.float64] | np.float64:
    r"""
    Computes the Gaussian Q-function. It is given by
    $$
        \mathrm{Q}(x) = \frac{1}{\sqrt{2\pi}} \int_x^\infty \mathrm{e}^{-u^2/2} \, \mathrm{d}u.
    $$

    Parameters:
        x (float | ArrayND[float]): The input to the function. May be any float or array of floats.

    Returns:
        y (SameAsInput): The value $y = \mathrm{Q}(x)$.

    Examples:
        >>> komm.qfunc(0.0)
        np.float64(0.5)

        >>> komm.qfunc([[-1.0], [0.0], [1.0]])
        array([[0.84134475],
               [0.5       ],
               [0.15865525]])
    """
    result = np.vectorize(_qfunc)(x)
    return np.float64(result) if np.isscalar(x) else result


def qfuncinv(y: npt.ArrayLike) -> npt.NDArray[np.float64] | np.float64:
    r"""
    Computes the inverse Gaussian Q-function.

    Parameters:
        y (float | ArrayND[float]): The input to the function. Should be a float or array of floats in the real interval $[0, 1]$.

    Returns:
        x (SameAsInput): The value $x = \mathrm{Q^{-1}}(y)$.

    Examples:
        >>> komm.qfuncinv(0.5)
        np.float64(0.0)

        >>> komm.qfuncinv([[0.841344746], [0.5], [0.158655254]])
        array([[-1.],
               [ 0.],
               [ 1.]])
    """
    result = np.vectorize(_qfuncinv)(y)
    return np.float64(result) if np.isscalar(y) else result


def logcosh(x: npt.ArrayLike) -> np.float64:
    # https://stackoverflow.com/a/57786270/
    s = np.sign(x) * x
    p = np.exp(-2 * s)
    return s + np.log1p(p) - np.log(2)



from typing import Any

import numpy as np
import numpy.typing as npt

from ..._error_control_convolutional.TerminatedConvolutionalCode import (
    TerminatedConvolutionalCode,
)
from ..._finite_state_machine.FiniteStateMachine import MetricFunction
from ..._util import unpack
from ..registry import RegistryBlockDecoder


def decode_viterbi(
    code: TerminatedConvolutionalCode,
    r: npt.ArrayLike,
    metric_function: MetricFunction[Any],
) -> npt.NDArray[np.int_]:
    if code.mode == "tail-biting":
        raise NotImplementedError("Viterbi algorithm not implemented for 'tail-biting'")

    k0 = code.convolutional_code.num_input_bits
    n0 = code.convolutional_code.num_output_bits
    mu = code.convolutional_code.memory_order
    fsm = code.convolutional_code.finite_state_machine()

    initial_metrics = np.full(fsm.num_states, fill_value=np.inf)
    initial_metrics[0] = 0.0

    xs_hat, final_metrics = fsm.viterbi(
        observed_sequence=np.reshape(r, shape=(-1, n0)),
        metric_function=metric_function,
        initial_metrics=initial_metrics,
    )

    if code.mode == "direct-truncation":
        s_hat = np.argmin(final_metrics)
        x_hat = xs_hat[:, s_hat]
    else:  # code.mode == "zero-termination"
        x_hat = xs_hat[:, 0][:-mu]

    u_hat = unpack(x_hat, width=k0)
    return u_hat


def decode_viterbi_hard(
    code: TerminatedConvolutionalCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    def metric_function(y: int, z: float) -> float:
        return np.count_nonzero(code.cache_bit[y] != z)

    return decode_viterbi(code, r, metric_function)


RegistryBlockDecoder.register(
    "viterbi-hard",
    {
        "description": "Viterbi (hard-decision)",
        "decoder": decode_viterbi_hard,
        "type_in": "hard",
        "type_out": "hard",
        "target": "message",
    },
)


def decode_viterbi_soft(
    code: TerminatedConvolutionalCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    def metric_function(y: int, z: int) -> float:
        return np.dot(code.cache_bit[y], z)

    return decode_viterbi(code, r, metric_function)


RegistryBlockDecoder.register(
    "viterbi-soft",
    {
        "description": "Viterbi (soft-decision)",
        "decoder": decode_viterbi_soft,
        "type_in": "soft",
        "type_out": "hard",
        "target": "message",
    },
)
